from collections import deque

class DAG:
    """Directed acyclic graph structure to manage pipeline task dependecies."""

    def __init__(self):
        self.graph = {}


    def add(self, node, points_to=None):
        """Add new task not to the graph, specify optional 'points_to' parameter."""
        if node not in self.graph:
            self.graph[node] = []

        if points_to:
            if points_to not in self.graph:
                self.graph[points_to] = []
            self.graph[node].append(points_to) # todo: need to make sure not to add duplicates

        # if sorted tasks and original graph lengths there must be a cycle 
        if len(self.sort()) != len(self.graph):
            raise Exception('A cycle is detected in the graph')


    def sort(self):
        """Sort all the task nodes based on the dependencies."""
        self.in_degrees()

        nodes_to_visit = deque()
        for node, pointers in self.degrees.items():
            # find all root nodes
            if pointers == 0:
                nodes_to_visit.append(node)

        sorted_nodes = []

        while nodes_to_visit:
            node = nodes_to_visit.popleft()
            for pointer in self.graph[node]:
                self.degrees[pointer] -= 1
                if self.degrees[pointer] == 0:
                    nodes_to_visit.append(pointer)
            sorted_nodes.append(node)

        return sorted_nodes


    def in_degrees(self):
        """Determing number of in-coming edges for each task node."""
        self.degrees = {}

        for node in self.graph:
            if node not in self.degrees:
                self.degrees[node] = 0
            for pointed in self.graph[node]:
                if pointed not in self.degrees:
                    self.degrees[pointed] = 0
                self.degrees[pointed] += 1





import numpy as np
import textwrap
import copy
import warnings
from scipy.stats import norm, beta
from shutil import get_terminal_size

from mbnpy.variable import Variable, findnth
from mbnpy import utils
from mbnpy.extern.tabulate import tabulate

class Cpm(object):
    '''Defines the conditional probability matrix (CPM).
    CPM plays the same role as CPT in conventional Bayesian networks.
    Ref: Byun et al. (2019). Matrix-based Bayesian Network for
    efficient memory storage and flexible inference.
    Reliability Engineering & System Safety, 185, 533-545.

    Attributes:
        variables (list): list of instances of Variable objects.
        no_child (int): number of child nodes.
        C (array_like): event matrix.
        p (array_like): probability vector related to C.
        Cs (array_like): event matrix of samples.
        q (array_like): sampling probability vector related to Cs.
        sample_idx (array_like): sample index vector related to Cs.
        ps (array_like): probability vector of samples.
        Sample weights are calculated as ps/q.

    Notes:
        C and p have the same number of rows.
        Cs, q, sample_idx, and ps have the same number of rows.
    '''

    def __init__(self, variables, no_child, C=[], p=[], Cs=[], q=[], ps=[], sample_idx=[]):

        self.variables = variables
        self.no_child = no_child
        self.C = C
        self.p = p
        self.Cs = Cs
        self.q = q
        self.ps = ps
        self.sample_idx = sample_idx

    # Magic methods
    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        if isinstance(other, Cpm):
            return (
                self._name == other._name and
                self._variables == other._variables and
                self._no_child == other._no_child and
                self._C == other._C and
                self._p == other._p and
                self._Cs == other._Cs and
                self._q == other._q and
                (
                    (not self._ps or not other._ps or self._ps == other._ps) and
                    (not self._sample_idx or not other._sample_idx or
                     self._sample_idx == other._sample_idx)
                )
            )
        else:
            return False

    def __repr__(self):

        if len(self.get_names()) > 7:
            names = self.get_names()[min(3, self._no_child)] + '...' + self.get_names()[-max(3, no_child-1):]
        else:
            names = self.get_names()[self._no_child:]

        _str = ", ".join(self.get_names()[0:self._no_child]) + ' | ' + ", ".join(names)

        return f"<CPM representing P({_str}) at {hex(id(self))}>"


    def __str__(self):

        #if len(self.get_names()) > 7:
        #    names = self.get_names()[min(3, self._no_child)] + '...' + self.get_names()[-max(3, no_child-1):]
        #else:
        names = self.get_names()[self._no_child:]

        _str = ", ".join(self.get_names()[:self._no_child]) + ' | ' + ", ".join(names)

        """
        if len(self.get_names()) > 7:
            names = self.get_names()[1:4] + '...' + self.get_names()[-3:]
        else:
            names = self.get_names()[1:]
        _str = self.get_names()[0] + ' | ' + ", ".join(names)
        """
        header = self.get_names() + ['p']
        c_p = np.append(self.C, self.p, axis=1)
        cpm_str = tabulate(c_p, headers=header, tablefmt='grid')

        # replace | with || 
        cpm_str = self._truncate_strtable(cpm_str)
        cpm_str = self._condition_strtable(cpm_str)

        details = [
            f"<CPM representing P({_str}) at {hex(id(self))}>",
            f"{cpm_str}",
        ]
        """
        if self._Cs.size:
            details.append(f"Cs={self._Cs},")
        if self._q.size:
            details.append(f"q={self._q},")
        if self._ps.size:
            details.append(f"ps={self._ps},")
        if self._sample_idx.size:
            details.append(f"sample_idx={self._sample_idx},")
        details.append(")")
        """
        return "\n".join(details)


    def _condition_strtable(self, cpm_str):

        list_rows_str = cpm_str.split("\n")

        idx_child = findnth(list_rows_str[1], '|', self.no_child)

        idx_p = findnth(list_rows_str[1], '|', list_rows_str[1].count('|') - 2)

        new_cpm_str = [list_rows_str[0]]
        for i, line in enumerate(list_rows_str[1:-1], 1):
            if i % 2 == 1:
                new_line = line[:idx_child-1] + ' [' + line[idx_child+1:idx_p-1] + ' ]' + line[idx_p+1:]
            else:
                new_line = line

            new_cpm_str.append(new_line)
        new_cpm_str.append(list_rows_str[-1])

        return "\n".join(new_cpm_str)


    def _truncate_strtable(self, cpm_str):
        terminal_width, terminal_height = get_terminal_size()

        list_rows_str = cpm_str.split("\n")

        table_width, table_height = len(list_rows_str[0]), len(list_rows_str)

        colstr_i = np.array(
            [pos for pos, char in enumerate(list_rows_str[0]) if char == "+"]
        )

        if table_width > terminal_width:
            half_width = terminal_width // 2 - 3

            left_i = colstr_i[colstr_i < half_width][-1]
            right_i = colstr_i[(table_width - colstr_i) < half_width][0]

            new_cpm_str = []
            for temp_row_str in list_rows_str:
                left = temp_row_str[: left_i + 1]
                right = temp_row_str[right_i:]
                if temp_row_str[left_i] == "+":
                    joiner = "-----"
                else:
                    joiner = " ... "
                new_cpm_str.append(left + joiner + right)

            cpm_str = "\n".join(new_cpm_str)

        if table_height > terminal_height:

            half_height = terminal_height // 2 - 3

            list_rows_str = cpm_str.split("\n")
            mid = findnth(list_rows_str[0], '+', 1)

            if half_height % 2:
                first_chunk = half_height + 2
                mid_chunk = half_height + 1
                last_chunk = -half_height + 2
            else:
                first_chunk = half_height + 1
                mid_chunk = half_height
                last_chunk = -half_height + 1

            new_cpm_str = list_rows_str[:first_chunk]
            no_cross = list_rows_str[mid_chunk].count('+')
            idx_cross = [findnth(list_rows_str[mid_chunk], '+', i) for i in range(no_cross)]
            mid_cross = [(x + y)//2 for x, y in zip(idx_cross[:-1], idx_cross[1:])]
            re_line = list(table_width * ' ')
            for i in idx_cross:
                re_line[i] = '|'
            for i in mid_cross:
                re_line[i] = ':'
            re_line = ''.join(re_line)
            new_cpm_str.append(re_line)

            [new_cpm_str.append(x) for x in list_rows_str[last_chunk:]]

            cpm_str = "\n".join(new_cpm_str)

        return cpm_str


    # Properties
    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, value):
        assert isinstance(value, list), 'variables must be a list of Variable'
        assert all([isinstance(x, Variable) for x in value]), 'variables must be a list of Variable'
        self._variables = value

    @property
    def no_child(self):
        return self._no_child

    @no_child.setter
    def no_child(self, value):
        assert isinstance(value, (int, np.int32, np.int64)), 'no_child must be a numeric scalar'
        assert value <= len(self.variables), f'no_child must be less than or equal to the number of variables: {value}, {len(self.variables)}'
        self._no_child = value

    @property
    def C(self):
        return self._C

    @C.setter
    def C(self, value):
        if isinstance(value, list):
            value = np.array(value, dtype=np.int32)

        if value.size:
            assert value.dtype in (np.dtype('int64'), np.dtype('int32')), f'Event matrix C must be a numeric matrix: {value}'

            if value.ndim == 1:
                value.shape = (len(value), 1)
            else:
                assert value.shape[1] == len(self._variables), 'C must have the same number of columns as that of variables'

            max_C = np.max(value, axis=0, initial=0)
            max_var = [2**len(x.values)-1 for x in self._variables]
            assert all(max_C <= max_var), f'check C matrix: {max_C} vs {max_var}'

        self._C = value

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, value):
        if isinstance(value, list):
            value = np.array(value)[:, np.newaxis]

        if value.ndim == 1:
            value.shape = (len(value), 1)

        if value.size:
            assert len(value) == self._C.shape[0], 'p must have the same length as the number of rows in C'

            all(isinstance(y, (float, np.float32, np.float64, int, np.int32, np.int64)) for y in value), 'p must be a numeric vector'

        self._p = value

    @property
    def Cs(self):
        return self._Cs

    @Cs.setter
    def Cs(self, value):
        if isinstance(value, list):
            value = np.array(value, dtype=np.int32)

        if value.size:
            if value.ndim == 1: # event matrix for samples
                value.shape = (len(value), 1)
            else:
                assert value.shape[1] == len(self._variables), 'Cs must have the same number of columns as the number of variables'

            max_Cs = np.max(value, axis=0, initial=0)
            max_var_basic = [len(x.values) for x in self.variables]
            assert all(max_Cs <= max_var_basic), f'check Cs matrix: {max_Cs} vs {max_var_basic}'

        self._Cs = value

    @property
    def q(self):
        return self._q

    @q.setter
    def q(self, value):
        if isinstance(value, list):
            value = np.array(value)[:, np.newaxis]

        if value.ndim == 1:
            value.shape = (len(value), 1)

        if value.size and self._Cs.size:
            assert len(value) == self._Cs.shape[0], 'q must have the same length as the number of rows in Cs'

        if value.size:
            all(isinstance(y, (float, np.float32, np.float64, int, np.int32, np.int64)) for y in value), 'q must be a numeric vector'

        self._q = value

    @property
    def ps(self):
        return self._ps

    @ps.setter
    def ps(self, value):
        if isinstance(value, list):
            value = np.array(value)[:, np.newaxis]

        if value.ndim == 1:
            value.shape = (len(value), 1)

        if value.size and self._Cs.size:
            assert len(value) == self._Cs.shape[0], 'ps must have the same length as the number of rows in Cs'

            all(isinstance(y, (float, np.float32, np.float64, int, np.int32, np.int64)) for y in value), 'p must be a numeric vector'

        self._ps = value

    @property
    def sample_idx(self):
        return self._sample_idx

    @sample_idx.setter
    def sample_idx(self, value):
        if isinstance(value, list):
            value = np.array(value)[:, np.newaxis]

        if value.ndim == 1:
            value.shape = (len(value), 1)

        if value.size and self._Cs.size:
            assert len(value) == self._Cs.shape[0], 'sample_idx must have the same length as the number of rows in Cs'

            all(isinstance(y, (float, np.float32, np.float64, int, np.int32, np.int64)) for y in value), 'p must be a numeric vector'

        self._sample_idx = value


    def product(self, M):
        """
        Returns an instance of Cpm
        M: instance of Cpm
        var: a dict of instances of Variable
        """

        assert isinstance(M, Cpm), f'M should be an instance of Cpm'

        ch_names1 = [x.name for x in self.variables[:self.no_child]]
        ch_names2 = [x.name for x in M.variables[:M.no_child]]
        check = set(ch_names1).intersection(ch_names2)
        assert not bool(check), f'PMFs must not have common child nodes: {ch_names1}, {ch_names2}'

        if self.C.size or self.Cs.size:
            idx_vars, _ = ismember(self.variables, M.variables)
            prod_vars = M.variables + get_value_given_condn(self.variables, flip(idx_vars))

            new_child = self.variables[:self.no_child] + M.variables[:M.no_child]

            new_parent = self.variables[self.no_child:] + M.variables[M.no_child:]
            new_parent, _ = setdiff(new_parent, new_child)

            if new_parent:
                new_vars = new_child + new_parent
            else:
                new_vars = new_child
            _, idx_vars2 = ismember(new_vars, prod_vars)

            Mprod = Cpm(variables=new_vars,
                        no_child = len(new_child))
        else:
            Mprod = M

        # Product starts here
        if self._C.size:
            if M._C.size:
                Cnew, pnew = C_prod_C(self, M, new_vars)
                Mprod.C, Mprod.p = Cnew, pnew

                if M._Cs.size and self._Cs.size:
                    Cs_new, q_new, ps_new, sample_idx_new = Cs_prod_Cs(
                    self.Cs, M.Cs, self.variables, M.variables,
                    self.q, M.q, self.ps, M.ps, self.sample_idx, M.sample_idx, new_vars)

                    Mprod.Cs, Mprod.q = Cs_new, q_new
                    Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new

                elif M._Cs.size and not self._Cs.size:
                    Cs_new, q_new, ps_new, sample_idx_new = C_prod_Cs(self, M, new_vars)
                    Mprod.Cs, Mprod.q = Cs_new, q_new
                    Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new

                elif self._Cs.size and not M._Cs.size:
                    Cs_new, q_new, ps_new, sample_idx_new = C_prod_Cs(M, self, new_vars)
                    Mprod.Cs, Mprod.q = Cs_new, q_new
                    Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new

            elif M._Cs.size and not self._Cs.size:
                Cs_new, q_new, ps_new, sample_idx_new = C_prod_Cs(self, M, new_vars)
                Mprod.Cs, Mprod.q = Cs_new, q_new
                Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new

            elif self._Cs.size:
                Cs_new1, q_new1, ps_new1, sample_idx_new1 = C_prod_Cs(self, M, new_vars)
                Cs_new2, q_new2, ps_new2, sample_idx_new2 = Cs_prod_Cs(
                    self.Cs, M.Cs, self.variables, M.variables,
                    self.q, M.q, self.ps, M.ps, self.sample_idx, M.sample_idx, new_vars)

                Cs_new = np.vstack([Cs_new1, Cs_new2])
                q_new = np.concatenate([q_new1, q_new2])
                ps_new = np.concatenate([ps_new1, ps_new2])
                sample_idx_new = np.concatenate([sample_idx_new1.flatten(), sample_idx_new2.flatten()])

                # Sort all sample arrays based on sample_idx
                sort_idx = np.argsort(sample_idx_new)
                Cs_new = Cs_new[sort_idx]
                q_new = q_new[sort_idx]
                ps_new = ps_new[sort_idx]
                sample_idx_new = sample_idx_new[sort_idx]

                Mprod.Cs, Mprod.q, = Cs_new.astype(int), q_new
                Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new.astype(int)

        else:
            if M._C.size and M._Cs.size:
                Cs_new1, q_new1, ps_new1, sample_idx_new1 = C_prod_Cs(M, self, new_vars)
                Cs_new2, q_new2, ps_new2, sample_idx_new2 = Cs_prod_Cs(
                    M.Cs, self.Cs, M.variables, self.variables,
                    M.q, self.q, M.ps, self.ps, M.sample_idx, self.sample_idx, new_vars)

                Cs_new = np.vstack([Cs_new1, Cs_new2])
                q_new = np.vstack([q_new1, q_new2])
                ps_new = np.concatenate([ps_new1, ps_new2])
                sample_idx_new = np.concatenate([sample_idx_new1, sample_idx_new2])

                sort_idx = np.argsort(sample_idx_new.flatten())
                Cs_new = Cs_new[sort_idx]
                q_new = q_new[sort_idx]
                ps_new = ps_new[sort_idx]

                Mprod.Cs, Mprod.q = Cs_new, q_new
                Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new

            elif M._C.size:
                Cs_new, q_new, ps_new, sample_idx_new = C_prod_Cs(M, self, new_vars)
                Mprod.Cs, Mprod.q = Cs_new, q_new
                Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new

            elif M._Cs.size:
                Cs_new, q_new, ps_new, sample_idx_new = Cs_prod_Cs(
                    self.Cs, M.Cs, self.variables, M.variables,
                    self.q, M.q, self.ps, M.ps, self.sample_idx, M.sample_idx, new_vars)

                Mprod.Cs, Mprod.q, = Cs_new, q_new
                Mprod.ps, Mprod.sample_idx = ps_new, sample_idx_new

        if Mprod.Cs.size:
            Mprod.Cs = Mprod.Cs.astype(int)

        return  Mprod

    def _get_Cnew(self, M, new_vars):
        """
        For internal use in "product" method
        """

        new_vars_tf1, new_vars_idx1 = ismember( new_vars, self.variables )
        new_vars_tf2, new_vars_idx2 = ismember( new_vars, M.variables )

        n_row1 = len(self.C)
        n_row2 = len(M.C)

        Bst_dict = {}
        for k, v in enumerate(new_vars):

            if new_vars_tf1[k] and new_vars_tf2[k]: # common variable to product

                n_val = len(v.values)

                Bvec1 = np.zeros((n_row1, n_val), dtype=int)
                Bvec2 = np.zeros((n_row2, n_val), dtype=int)

                C1 = self.C[:, new_vars_idx1[k]]
                for i, b in enumerate(C1):
                    st = list( v.B[ b ] )
                    Bvec1[i, st] = 1
                Bvec1 = np.tile(Bvec1, (n_row2, 1, 1))

                C2 = M.C[:, new_vars_idx2[k]]
                for i, b in enumerate(C2):
                    st = list( v.B[ b ] )
                    Bvec2[i, st] = 1
                Bvec2 = np.tile(Bvec2[:, np.newaxis, :], (1, n_row1, 1))

                Bvec = Bvec1 * Bvec2
                Bst = v.get_Bst_from_Bvec( Bvec )

                Bst_dict[v.name] = Bst

            elif new_vars_tf1[k]: # it is in self only
                C1 = self.C[:, new_vars_idx1[k]]
                Bst = np.tile(C1.T, (n_row2, 1, 1))
                Bst = np.squeeze(Bst)

                Bst_dict[v.name] = Bst

            else: # it is in M only
                C2 = M.C[:, new_vars_idx2[k]]
                Bst = np.repeat(C2.reshape(-1,1), n_row1, axis=1)

                Bst_dict[v.name] = Bst

        Cnew_list = []
        for v in new_vars:
            Cnew_list.append( Bst_dict[v.name].T.reshape(-1) )

        # Cnew with -1
        # -1 elements mean incompatible rows between the two CPMs
        Cnew_with_minus1 = np.column_stack( Cnew_list ) 

        return Cnew_with_minus1

    def _C_prod_C(self, M, new_vars):
        """
        For internal use in "product" method
        """

        new_vars_tf1, new_vars_idx1 = ismember( new_vars, self.variables )
        new_vars_tf2, new_vars_idx2 = ismember( new_vars, M.variables )

        n_row1 = len(self.C)
        n_row2 = len(M.C)

        Bst_dict = {}
        for k, v in enumerate(new_vars):

            if new_vars_tf1[k] and new_vars_tf2[k]: # common variable to product

                n_val = len(v.values)

                Bvec1 = np.zeros((n_row1, n_val), dtype=int)
                Bvec2 = np.zeros((n_row2, n_val), dtype=int)

                C1 = self.C[:, new_vars_idx1[k]]
                for i, b in enumerate(C1):
                    st = list( v.B[ b ] )
                    Bvec1[i, st] = 1
                Bvec1 = np.tile(Bvec1, (n_row2, 1, 1))

                C2 = M.C[:, new_vars_idx2[k]]
                for i, b in enumerate(C2):
                    st = list( v.B[ b ] )
                    Bvec2[i, st] = 1
                Bvec2 = np.tile(Bvec2[:, np.newaxis, :], (1, n_row1, 1))

                Bvec = Bvec1 * Bvec2
                Bst = v.get_Bst_from_Bvec( Bvec )

                Bst_dict[v.name] = Bst

            elif new_vars_tf1[k]: # it is in self only
                C1 = self.C[:, new_vars_idx1[k]]
                Bst = np.tile(C1.T, (n_row2, 1, 1))
                Bst = np.squeeze(Bst)

                Bst_dict[v.name] = Bst

            else: # it is in M only
                C2 = M.C[:, new_vars_idx2[k]]
                Bst = np.repeat(C2.reshape(-1,1), n_row1, axis=1)

                Bst_dict[v.name] = Bst

        Cnew_list = []
        for v in new_vars:
            Cnew_list.append( Bst_dict[v.name].T.reshape(-1) )
        Cnew = np.column_stack( Cnew_list )

        pnew = np.repeat(self.p, n_row2) * np.tile(M.p, (n_row1, 1)).flatten()

        mask = np.sum( Cnew < 0, axis=1 ) < 1

        Cnew = Cnew[mask]
        pnew = pnew[mask]

        return Cnew, pnew

    def get_variables(self, item):

        if isinstance(item, str):
            return [x for x in self.variables if x.name == item][0]
        elif isinstance(item, list):
            return [self.get_variables(y) for y in item]


    def get_names(self):
        return [x.name for x in self.variables]


    def get_subset(self, row_idx, flag=True, isC=True):
        """
        Returns the subset of Cpm

        Parameters
        ----------
        row_idx: array like
        flag: boolean
            default True, 0 if exclude row_idx
        isC: boolean
            if True, C and p are reduced; if False, Cs, q, ps, sample_idx are.
        """

        assert flag in (0, 1)
        assert isC in (0, 1)

        if isC:
            if flag:
                assert set(row_idx).issubset(range(self.C.shape[0]))
            else:
                # select row excluding the row_index
                row_idx, _ = setdiff(range(self.C.shape[0]), row_idx)

            if self.p.size:
                p_sub = self.p[row_idx]
            else:
                p_sub = []

            Mnew = Cpm(variables=self.variables,
                       no_child=self.no_child,
                       C=self.C[row_idx,:],
                       p=p_sub,
                       Cs = self.Cs,
                       q = self.q,
                       ps = self.ps,
                       sample_idx = self.sample_idx)

        else:
            if flag:
                assert set(row_idx).issubset(range(self.Cs.shape[0]))
            else:
                # select row excluding the row_index
                row_idx, _ = setdiff(range(self.Cs.shape[0]), row_idx)

            if self.q.size:
                q_sub = self.q[row_idx]
            else:
                q_sub = []

            if self.ps.size:
                ps_sub = self.ps[row_idx]
            else:
                ps_sub = []

            if self.sample_idx.size:
                si_sub = self.sample_idx[row_idx]
            else:
                si_sub = []

            Mnew = Cpm(variables=self.variables,
                       no_child=self.no_child,
                       Cs=self.Cs[row_idx,:],
                       q=q_sub,
                       ps=ps_sub,
                       sample_idx = si_sub)

        return Mnew


    def get_means(self, names):
        """
        Get means of variables in names
        INPUT:
        names: a list of names
        OUTPUT:
        means: a list of means (the same order as names)
        """
        assert isinstance(names, list), 'names should be a list'
        assert len(set(names))==len(names), f'names has duplicates: {names}'

        idx = [self.get_names().index(x) for x in names]

        return [(self.C[:, i]*self.p[:, 0]).sum() for i in idx]


    def iscompatible(self, M, composite_state=True):
        """
        Returns a boolean list (n,)

        Parameters
        ----------
        M: instance of Cpm for compatibility check
        composite_state: False if the same rows are returned;
            True if composite states and basic state can be
            considered compatible if they are.
        """

        assert isinstance(M, Cpm), f'M should be an instance of Cpm'

        assert ( (M.C.shape[0] == 1) and (not M.Cs.size) ) or ( (M.Cs.shape[0] == 1) and (not M.C.size) ), 'C/Cs must be a single row'

        if self.C.size and M.C.size:
            is_cmp = iscompatible(self.C, self.variables, M.variables, M.C[0], composite_state)

        else:
            is_cmp = []

        return is_cmp


    def get_col_ind(self, names: list[str]):
        """
        INPUT:
        names: a list of variable names
        OUTPUT:
        idx: a list of column indices of v_names
        """

        assert isinstance(names, list), f'names should be a list'

        assert len(set(names)) == len(names), f'names has duplicates: {names}'

        return [self.get_names().index(x) for x in names]

    def get_col(self, names):

        assert isinstance(names, list), f'names should be a list'

        assert len(set(names)) == len(names), f'names has duplicates: {names}'

        inds = [self.get_names().index(x) for x in names]

        C = np.zeros((len(self.C), len(names)), dtype=int)
        for i, col_ind in enumerate(inds):
            C[:,i] = self.C[:,col_ind]

        return C



    def merge(self, M):

        assert isinstance(M, Cpm), f'M should be an instance of Cpm'
        assert self.variables == M.variables, 'must have the same scope and order'

        new_cpm = copy.copy(self)

        cs = self.C.tolist()

        for cx, px in zip(M.C.tolist(), M.p.tolist()):
            try:
                new_cpm.p[cs.index(cx)] += px
            except IndexError:
                new_cpm.C = np.vstack((new_cpm.C, cx))
                new_cpm.p = np.vstack((new_cpm.p, px))

        new_cpm.Cs = np.vstack((self.Cs, M.Cs))
        new_cpm.q = np.vstack((self.q, M.q))
        new_cpm.ps = np.vstack((self.ps, M.ps))
        new_cpm.sample_idx = np.vstack((self.sample_idx, M.sample_idx))

        return new_cpm


    def sum(self, variables, flag=True):
        """
        Returns instance of Cpm with based on Sum over CPMs.

        Parameters
        ----------
        variables: list of variables or names of variables
        flag: boolean
            1 (default) - sum out variables, 0 - leave only variables
        """

        assert isinstance(variables, list), 'variables should be a list'
        if variables and isinstance(variables[0], str):
            variables = self.get_variables(variables)

        if flag and set(self.variables[self.no_child:]).intersection(variables):
            print('Parent nodes are NOT summed up')

        if flag:
            vars_rem, vars_rem_idx = setdiff(self.variables[:self.no_child], variables)

        else:
            # FIXME
            _, vars_rem_idx = ismember(variables, self.variables[:self.no_child])
            vars_rem_idx = get_value_given_condn(vars_rem_idx, vars_rem_idx)
            vars_rem_idx = sorted(vars_rem_idx)
            vars_rem = [self.variables[x] for x in vars_rem_idx]

        no_child_sum = len(vars_rem)

        if self.variables[self.no_child:]:
            vars_rem += self.variables[self.no_child:]
            vars_rem_idx += list(range(self.no_child, len(self.variables)))

        _variables = [self.variables[i] for i in vars_rem_idx]

        M = Cpm(variables=_variables,
                C=self.C[:, vars_rem_idx],
                no_child=len(vars_rem_idx),
                p = self.p)

        Csum = np.array([], dtype=int).reshape(0, len(vars_rem))
        psum = []

        while M.C.size:

            Mc = M.get_subset([0])
            is_cmp = iscompatible(M.C, vars_rem, vars_rem, Mc.C, composite_state=False)

            Csum = np.vstack((Csum, Mc.C[0]))

            if M.p.size:
                psum.append(np.sum(M.p[is_cmp]))

            M = M.get_subset(np.where(is_cmp)[0], flag=0)

        Ms = Cpm(variables=vars_rem,
                 no_child=no_child_sum,
                 C=Csum,
                 p=np.reshape(psum, (-1, 1)))

        if self.Cs.size:

            Cs = self.Cs[:, vars_rem_idx].copy()

            Ms.Cs = Cs.astype(int)
            Ms.q = self.q.copy()
            Ms.ps = self.ps.copy()
            Ms.sample_idx = self.sample_idx.copy()

        return Ms


    def get_prob(self, var_inds, var_states, flag=True):

        assert isinstance(var_inds, list), 'var_inds should be a list'

        if var_inds and isinstance(var_inds[0], str):
            var_inds = self.get_variables(var_inds)

        assert isinstance(var_states, (list, np.ndarray)), 'var_states should be an array'

        assert len(var_inds) == len(var_states), f'"var_inds" {var_inds} and "var_states" {var_states} must have the same length.'

        assert flag in (0, 1), 'Operation flag must be either 1 (keeping given row indices default) or 0 (deleting given indices)'

        _var_states = []
        for i, x in enumerate(var_states):
            if isinstance(x, str):
                assert x in var_inds[i].values, f'{x} not in {var_inds[i].values}'
                _var_states.append(var_inds[i].values.index(x))

        if _var_states:
            var_states = _var_states[:]

        if isinstance(var_states, list):
            var_states = np.array(var_states)
            var_states = np.reshape(var_states, (1, -1))

        Mc = Cpm(variables=var_inds,
                 no_child=len(var_inds),
                 C=var_states,
                 p=np.empty(shape=(var_states.shape[0], 1)))

        is_compat = self.iscompatible(Mc, composite_state=True)
        idx = np.where(is_compat)[0]
        Msubset = self.get_subset(idx, flag)

        return Msubset.p.sum()


    def condition(self, cnd_vars, cnd_states):
        """
        Returns a list of instance of Cpm

        Parameters
        ----------
        cnd_vars: a list of variables to be conditioned
        cnd_states: a list of the states to be conditioned
        """

        assert isinstance(cnd_vars, (list, np.ndarray)), 'invalid cnd_vars'

        if isinstance(cnd_vars, np.ndarray):
            cnd_vars = cnd_vars.tolist()

        if cnd_vars and isinstance(cnd_vars[0], str):
            cnd_vars = self.get_variables(cnd_vars)

        assert isinstance(cnd_states, (list, np.ndarray)), 'invalid cnd_vars'

        if isinstance(cnd_states, np.ndarray):
            cnd_states = cnd_states.tolist()

        if cnd_states and isinstance(cnd_states[0], str):
            cnd_states = [x.values.index(y) for x, y in zip(cnd_vars, cnd_states)]

        Mx = copy.deepcopy(self)

        is_cmp = iscompatible(Mx.C, Mx.variables, cnd_vars, cnd_states)

        C = Mx.C[is_cmp, :].copy()

        _, idx_cnd = ismember(cnd_vars, Mx.variables)
        _, idx_vars = ismember(Mx.variables, cnd_vars)

        Ccond = np.zeros_like(C)
        not_idx_vars = flip(idx_vars)

        if C.size:
            Ccond[:, not_idx_vars] = get_value_given_condn(C, not_idx_vars)

        cnd_vars_m = get_value_given_condn(cnd_vars, idx_cnd)
        cnd_states_m = get_value_given_condn(cnd_states, idx_cnd)
        idx_cnd = get_value_given_condn(idx_cnd, idx_cnd)

        for cnd_var, state, idx in zip(cnd_vars_m, cnd_states_m, idx_cnd):
            try:
                B = cnd_var.B.copy()
            except NameError:
                raise(f'{cnd_var} is not defined')
            else:
                C1 = C[:, idx].copy().astype(int)
                check = [B[x].intersection(B[state]) for x in C1]
                Ccond[:, idx] = [x for x in ismember(check, B)[1]]

        Mx.C = Ccond.copy()

        if Mx.p.size:
            Mx.p = Mx.p[is_cmp]

        if Mx.Cs.size: # NB: ps is not properly updated if the corresponding instance is not in C.

            Cnew_with_minus1 = _get_Cnew(Mx.Cs, np.array([cnd_states]), Mx.variables, cnd_vars, Mx.variables)

            mask = np.sum( Cnew_with_minus1 < 0, axis=1 ) < 1
            Cs_new, ps_new = Cnew_with_minus1[mask], Mx.ps[mask]
            q_new, sample_idx_new = Mx.q[mask], Mx.sample_idx[mask]

            Mx.Cs, Mx.q, Mx.ps, Mx.sample_idx = Cs_new, q_new, ps_new, sample_idx_new

        return Mx


    def sort(self):
 
        idx = argsort(list(map(tuple, self.C)))
        self.C = self.C[idx, :]

        if self.p.size:
            self.p = self.p[idx]

        if self.sample_idx.size:
            idx_sample = argsort(self.sample_idx)

        if self.Cs.size:
            self.Cs = self.Cs[idx_sample, :]

        if self.q.size:
            self.q = self.q[idx_sample]

        if self.sample_idx.size:
            self.sample_idx = self.sample_idx[idx_sample]


    def get_prob_bnd(self, var_inds, var_states, flag=True, cvar_inds=None, cvar_states=None, cflag=True):

        prob1 = self.get_prob(var_inds, var_states, flag)
        prob_unk = 1.0 - np.sum(self.p) # Unknown probs
        prob1_bnd = [prob1, prob1 + prob_unk]

        if cvar_inds:
           prob2 =  self.get_prob(cvar_inds, cvar_states, cflag)
           prob2_bnd = [prob2, prob2 + prob_unk]
           prob_bnd = [prob1_bnd[0] / prob2_bnd[1], prob1_bnd[1] / prob2_bnd[0]]

        else:
           prob_bnd = prob1_bnd

        prob_bnd[1] = min(1, prob_bnd[1])

        return prob_bnd


    def get_prob_and_cov(self, var_inds, var_states, method='MLE', flag=True, nsample_repeat=0, conf_p=0.95):

        assert isinstance(nsample_repeat, int), 'nsample_repeat must be a nonnegative integer, representing if samples are repeated (to calculate c.o.v.)'

        prob_C = self.get_prob(var_inds, var_states, flag)

        if not nsample_repeat:
            n_round = 1
            nsamp = len(self.Cs)
        else:
            assert len(self.Cs) % nsample_repeat == 0, 'Given number of samples is not divided by given nsample_repeat'
            n_round = int(len(self.Cs) / nsample_repeat)
            nsamp = nsample_repeat

        prob_Cs = 0
        var = 0
        for i in range(n_round):

            row_range = range(i*nsamp, (i + 1)*nsamp)
            is_cmp = iscompatible(self.Cs[row_range,:], self.variables, var_inds, var_states)

            try:
                w = self.ps[row_range] / self.q[row_range]
            except IndexError:
                w = np.ones_like(self.q) # if ps is empty, assume ps is the same as q

            w_ori = w.copy() # weight before compatibility check
            if flag:
                w[~is_cmp] = 0
            else:
                w[is_cmp] = 0

            if method=='MLE':
                mean = w.sum() / nsamp
                prob_Cs += mean

                if np.allclose(w_ori, w_ori[0], atol=1e-4): # this is MCS
                    var1 = np.square(w_ori[0]) * (1 - mean) * mean / nsamp
                    var += var1[0]
                else:
                    var1 = np.square((w - mean) / nsamp)
                    var += var1.sum()

            elif method=='Bayesian':
                neff = len(w_ori)*w_ori.mean()**2 / (sum(x**2 for x in w_ori)/len(w_ori)) # effective sample size
                w_eff = w / w_ori.sum() *neff
                nTrue = w_eff.sum()

                # to avoid numerical errors
                if np.isnan(nTrue):
                    nTrue = 0.0
                if np.isnan(neff[0]):
                    neff[0] = 0.0

                try:
                    a, b = a + nTrue, b + (neff[0] - nTrue)
                except NameError:
                    prior = 0.01
                    a, b = prior + nTrue, prior + (neff[0] - nTrue)

        if method == 'MLE':
            prob = prob_C + (1 - self.p.sum()) * prob_Cs
            cov = (1 - self.p.sum()) * np.sqrt(var) / prob

            # confidence interval
            z = norm.pdf(1 - (1 - conf_p)*0.5) # for both tails
            prob_Cs_int = prob_Cs + z * np.sqrt(var) * np.array([-1, 1])
            cint = prob_C + (1 - self.p.sum()) * prob_Cs_int

        elif method == 'Bayesian':

            mean = a / (a + b)
            var = a*b / (a+b)**2 / (a+b+1)

            prob = prob_C + (1 - self.p.sum()) * mean
            cov = (1 - self.p.sum()) * np.sqrt(var) / prob

            low = beta.ppf(0.5*(1-conf_p), a, b)
            up = beta.ppf(1 - 0.5*(1-conf_p), a, b)
            cint = prob_C + (1 - self.p.sum()) * np.array([low, up])

        return prob, cov, cint


    def get_prob_and_cov_cond(self, var_inds, var_states, cvar_inds, cvar_states, nsample_repeat=0, conf_p=0.95):
        # Assuming beta distribution (i.e. Bayeisan inference)

        assert isinstance(nsample_repeat, int), 'nsample_repeat must be a nonnegative integer, representing if samples are repeated (to calculate c.o.v.)'

        if not nsample_repeat:
            n_round = 1
            nsamp = len(self.Cs)
        else:
            assert len(self.Cs) % nsample_repeat == 0, 'Given number of samples is not divided by given nsample_repeat'
            n_round = int(len(self.Cs) / nsample_repeat)
            nsamp = nsample_repeat

        for i in range(n_round):
            row_range = range(i*nsamp, (i + 1)*nsamp)

            try:
                w_ori = self.ps[row_range] / self.q[row_range] # weight before compatibility check
            except IndexError:
                w_ori = np.ones_like(self.q) # if ps is empty, assume ps is the same as q

            is_cmp1 = iscompatible(self.Cs[row_range,:], self.variables, var_inds, var_states)
            w1 = w_ori.copy()
            w1[~is_cmp1] = 0

            is_cmp2 = iscompatible(self.Cs[row_range,:], self.variables, cvar_inds, cvar_states)
            w2 = w_ori.copy()
            w2[~is_cmp2] = 0

            neff = len(w_ori)*w_ori.mean()**2 / (sum(x**2 for x in w_ori)/len(w_ori)) # effective sample size

            w1_eff = w1 / w_ori.sum() *neff
            nTrue1 = w1_eff.sum()

            w2_eff = w2 / w_ori.sum() *neff
            nTrue2 = w2_eff.sum()

            try:
                a1, b1 = a1 + nTrue1, b1 + (neff[0] - nTrue1)
                a2, b2 = a2 + (nTrue2-nTrue1), b2 + (neff[0] - (nTrue2-nTrue1))
            except NameError:
                prior = 0.01
                a1, b1 = prior + nTrue1, prior + (neff[0] - nTrue1)
                a2, b2 = prior + (nTrue2-nTrue1), prior + (neff[0] - (nTrue2-nTrue1))

        prob_C = self.get_prob(var_inds, var_states)
        prob_C_c = self.get_prob(cvar_inds, cvar_states)

        prob, std, cint = utils.get_rat_dist( prob_C, prob_C_c - prob_C, 1 - self.p.sum(), a1, a2, b1, b2, conf_p )

        cov = std/prob

        return prob, cov, cint

def C_prod_C(M1, M2, new_vars):
    """
    Product of two C matrices (i.e. not Cs matrices)
    """
    Cnew_with_minus1 = _get_Cnew(M1.C, M2.C, M1.variables, M2.variables, new_vars)

    n_row1, n_row2 = len(M1.C), len(M2.C)
    pnew = np.repeat(M1.p, n_row2) * np.tile(M2.p, (n_row1, 1)).flatten()

    mask = np.sum( Cnew_with_minus1 < 0, axis=1 ) < 1
    Cnew, pnew = Cnew_with_minus1[mask], pnew[mask]

    return Cnew, pnew

def C_prod_Cs(M1, M2, new_vars):
    """
    Product of C (M1) and Cs (M2) matrices
    """
    Cnew_with_minus1 = _get_Cnew(M1.C, M2.Cs, M1.variables, M2.variables, new_vars)

    n_row1, n_row2 = len(M1.C), len(M2.Cs)
    ps_new = np.repeat(M1.p, n_row2) * np.tile(M2.ps, (n_row1, 1)).flatten()
    sample_idx_new = np.tile(M2.sample_idx, (n_row1, 1))
    q_new = np.repeat(M1.p, n_row2) * np.tile(M2.q, (n_row1, 1)).flatten()

    mask = np.sum( Cnew_with_minus1 < 0, axis=1 ) < 1
    Cs_new, ps_new = Cnew_with_minus1[mask], ps_new[mask]
    q_new, sample_idx_new = q_new[mask], sample_idx_new[mask]

    # Sort rows by sample_idx for readability
    sorted_idx = np.argsort(sample_idx_new.flatten())
    Cs_new, q_new = Cs_new[sorted_idx], q_new[sorted_idx]
    ps_new, sample_idx_new = ps_new[sorted_idx], sample_idx_new[sorted_idx]

    return Cs_new, q_new, ps_new, sample_idx_new

def Cs_prod_Cs(Cs1, Cs2, vars1, vars2, q1, q2, ps1, ps2, sample_idx1, sample_idx2, new_vars):
    """
    Product of two Cs matrices
    """

    # Give warnings if there is a mismatch in sample indices between Cs1 and Cs2
    # Mismatching indices will be ignored
    unique_sidx_M1 = np.unique(sample_idx1)
    unique_sidx_M2 = np.unique(sample_idx1)

    only_in_M1 = unique_sidx_M1 - unique_sidx_M2
    only_in_M2 = unique_sidx_M2 - unique_sidx_M1

    if only_in_M1.size or only_in_M2.size:
        warnings.warn(
            f"Mismatch in unique sample indices between M1 and M2.\n"
            f"Indices only in M1: {sorted(only_in_M1)}\n"
            f"Indices only in M2: {sorted(only_in_M2)}",
            UserWarning
        )

    new_vars_tf1, new_vars_idx1 = ismember( new_vars, vars1 )
    new_vars_tf2, new_vars_idx2 = ismember( new_vars, vars2 )

    # Common variables
    com_vars_idx1, com_vars_idx2 = [], []
    for tf1, tf2, idx1, idx2 in zip(new_vars_tf1, new_vars_tf2, new_vars_idx1, new_vars_idx2):
        if tf1 and tf2:
            com_vars_idx1.append(idx1)
            com_vars_idx2.append(idx2)

    Cs_new = np.array([], dtype=int).reshape(0, len(new_vars))
    qs_new = np.array([])
    ps_new = np.array([])
    sample_idx_new = np.array([])

    # Mapping Cs_new's columns to Cs1's and Cs2's columns
    # Cs_col_map[0] is the index of Cs1's columns
    # Cs_col_map[1] is the index of Cs2's columns
    Cs_col_map = [[], []]
    for i, (tf1, tf2, idx1, idx2) in enumerate(zip(new_vars_tf1, new_vars_tf2, new_vars_idx1, new_vars_idx2)):
        if tf1:
            Cs_col_map[0].append([i, idx1])
        elif tf2:
            Cs_col_map[1].append([i, idx2])

    for c1, q1, ps1, sidx1 in zip(Cs1, q1, ps1, sample_idx1):
        row_idx2 = np.where(sample_idx2 == sidx1)[0]
        for i2 in row_idx2:
            if Cs2[i2][com_vars_idx2] == c1[com_vars_idx1]:
                cs_new = np.zeros((1, len(new_vars)))
                for i, idx in Cs_col_map[0]:
                    cs_new[0, i] = c1[idx]
                for i, idx in Cs_col_map[1]:
                    cs_new[0, i] = Cs2[i2][idx]

                Cs_new = np.vstack([Cs_new, cs_new])
                qs_new = np.append(qs_new, q1*q2[i2])
                ps_new = np.append(ps_new, ps1*ps2[i2])
                sample_idx_new = np.append(sample_idx_new, sidx1)

    Cs_new.astype(int)
    sample_idx_new.astype(int)

    return Cs_new, qs_new, ps_new, sample_idx_new


def _get_Cnew(C1, C2, vars1, vars2, new_vars):
    """
    For internal use in "product" method
    """

    new_vars_tf1, new_vars_idx1 = ismember( new_vars, vars1 )
    new_vars_tf2, new_vars_idx2 = ismember( new_vars, vars2 )

    n_row1 = len(C1)
    n_row2 = len(C2)

    Bst_dict = {}
    for k, v in enumerate(new_vars):

        if new_vars_tf1[k] and new_vars_tf2[k]: # common variable to product

            n_val = len(v.values)

            Bvec1 = np.zeros((n_row1, n_val), dtype=int)
            Bvec2 = np.zeros((n_row2, n_val), dtype=int)

            C1_ = C1[:, new_vars_idx1[k]]
            for i, b in enumerate(C1_):
                st = list( v.get_set(b) )
                Bvec1[i, st] = 1
            Bvec1 = np.tile(Bvec1, (n_row2, 1, 1))

            C2_ = C2[:, new_vars_idx2[k]]
            for i, b in enumerate(C2_):
                st = list( v.get_set(b) )
                Bvec2[i, st] = 1
            Bvec2 = np.tile(Bvec2[:, np.newaxis, :], (1, n_row1, 1))

            Bvec = Bvec1 * Bvec2
            Bst = v.get_Bst_from_Bvec( Bvec )

            Bst_dict[v.name] = Bst

        elif new_vars_tf1[k]: # it is in self only
            C1_ = C1[:, new_vars_idx1[k]]
            Bst = np.tile(C1_.T, (n_row2, 1, 1))
            Bst = np.squeeze(Bst)

            Bst_dict[v.name] = Bst

        else: # it is in M only
            C2_ = C2[:, new_vars_idx2[k]]
            Bst = np.repeat(C2_.reshape(-1,1), n_row1, axis=1)

            Bst_dict[v.name] = Bst

    Cnew_list = []
    for v in new_vars:
        Cnew_list.append( Bst_dict[v.name].T.reshape(-1) )

    # Cnew with -1
    # -1 elements mean incompatible rows between the two CPMs
    Cnew_with_minus1 = np.column_stack( Cnew_list ) 

    return Cnew_with_minus1


def argsort(seq):

    #http://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python/3382369#3382369
    return sorted(range(len(seq)), key=seq.__getitem__)


def ismember(A, B):
    """
    A: vector
    B: list
    return
     Lia: logical true and false where data in A is found in B
     Lib: the list (same length as A) of index of the first matching elment in B or False for non-matching element

    """

    if isinstance(A, np.ndarray) and (A.ndim > 1):

        assert A.shape[1] == np.array(B).shape[1]

        res = []
        for x in A:
            v = np.where((np.array(B) == x).all(axis=1))[0]
            if len(v):
                res.append(v[0])
            else:
                res.append(False)

    elif isinstance(A, list) and isinstance(B, list):

        res  = [B.index(x) if x in B else False for x in A]

    else:

        if isinstance(B, np.ndarray) and (B.ndim > 1):
            assert len(A) == B.shape[1]

        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)
            res  = [np.where(np.array(B)==x)[0].min()
                 if x in B else False for x in A]

    lia = [False if x is False else True for x in res]

    return lia, res


def setdiff(first, second):
    """
    matlab setdiff equivalent
    """
    second = set(second)
    first = list(dict.fromkeys(first))
    val = [item for item in first if item not in second]
    iv = [first.index(x) for x in val]
    return val, iv


def get_value_given_condn(A, condn):

    if isinstance(A, np.ndarray) and A.ndim==2 and A.shape[1] == len(condn):
        A = A.T
        val = np.array([x for (i, x) in zip(condn, A) if i is not False])
        if val.size:
            val = val.reshape(-1, A.shape[1]).T
    else:
        assert len(A) == len(condn), f'len of {A} is not equal to len of {condn}'
        val = [x for (i, x) in zip(condn, A) if i is not False]

    return val


def iscompatible(C, variables, check_vars, check_states, composite_state=True):
    """
    Returns a boolean list

    Parameters
    ----------
    C: np.ndarray
    variables: array_like
    check_vars: array_like list of Variable or string
    check_sates: array_like list of index or string
    composite_state: False if the same rows are returned;
        True if composite states and basic state can be
        considered compatible if they are. 
    """
    if check_vars and isinstance(check_vars[0], str):
        check_vars = [x for y in check_vars for x in variables if x.name == y]

    _, idx = ismember(check_vars, variables)
    check_vars = get_value_given_condn(check_vars, idx)

    if len(check_vars) > 0:
        check_states = get_value_given_condn(check_states, idx)
        idx = get_value_given_condn(idx, idx)

        C = C[:, idx].copy()

        for i, (variable, state) in enumerate(zip(check_vars, check_states)):

            if isinstance(state, str):
                state = variable.values.index(state)
                check_states[i] = state

        if not composite_state:
            is_cmp = np.all(C == check_states, axis=1).tolist()

        else:
            Cnew_with_minus1 = _get_Cnew(C, np.array([check_states]), check_vars, check_vars, check_vars)
            is_cmp = np.all(Cnew_with_minus1 >= 0, axis=1).tolist()
    else:
        is_cmp = np.ones(shape=C.shape[0], dtype=bool).tolist()

    return is_cmp


def flip(idx):
    """
    boolean flipped
    Any int including 0 will be flipped False
    """
    return [True if x is False else False for x in idx]


def product(cpms):
    """
    return an instance of Cpm

    cpms: a list or dict of instances of Cpm
    """
    assert isinstance(cpms, (list,  dict)), 'cpms should be a list or dict'

    if isinstance(cpms, dict):
        cpms = list(cpms.values())

    assert cpms, f'{cpms} is empty list'

    prod = cpms[0]
    for cx in cpms[1:]:
        prod = prod.product(cx)

    return prod


def get_prod(A, B):
    """
    A: matrix
    B: matrix
    """
    if len(A.shape) < 2:
        A = np.reshape(A, (A.shape[0], 1))

    if len(B.shape) < 1:
        B=np.reshape(B,(1,))

    assert A.shape[1] == B.shape[0]

    prod_sign = np.sign(A * B)
    prod_val = np.exp(np.log(np.abs(A)) + np.log(np.abs(B)))

    return prod_sign * prod_val






'''
Ji-Eun Byun (ji-eun.byun@glasgow.ac.uk)
Created: 27 Mar 2023

A small, hypothetical bridge system
'''
import pytest
import numpy as np
import pandas as pd
import networkx as nx
import socket
import matplotlib
import pdb
import copy

from scipy.stats import lognorm
from pathlib import Path
from math import isclose

np.set_printoptions(precision=3)
#pd.set_option.display_precision = 3

if 'gadi' in socket.gethostname():
    matplotlib.use('Agg')
else:
    matplotlib.use("TKAgg")
import matplotlib.pyplot as plt

from mbnpy import cpm, variable, config, branch, model, trans, inference


HOME = Path(__file__).absolute().parent


@pytest.fixture(scope='package')
def data_bridge():
    """
    system with 6 edges, 5 nodes
    (see graph.png)
    """

    data = {}

    # Network
    data['node_coords'] = {'n1': (-2, 3),
                   'n2': (-2, -3),
                   'n3': (2, -2),
                   'n4': (1, 1),
                   'n5': (0, 0)}

    data['arcs'] = {'e1': ['n1', 'n2'],
            'e2': ['n1', 'n5'],
            'e3': ['n2', 'n5'],
            'e4': ['n3', 'n4'],
            'e5': ['n3', 'n5'],
            'e6': ['n4', 'n5']}

    data['arcs_avg_kmh'] = {'e1': 40,
                    'e2': 40,
                    'e3': 40,
                    'e4': 30,
                    'e5': 30,
                    'e6': 20}

    data['frag'] = {'major': {'med': 60.0, 'std': 0.7},
            'urban' : {'med': 24.0, 'std': 0.7},
            'bridge': {'med': 1.1, 'std': 3.9},
            }

    data['arcs_type'] = {'e1': 'major',
             'e2': 'major',
             'e3': 'major',
             'e4': 'urban',
             'e5': 'bridge',
             'e6': 'bridge'}

    data['var_ODs'] = {'od1': ('n5', 'n1'),
               'od2': ('n5', 'n2'),
               'od3': ('n5', 'n3'),
               'od4': ('n5', 'n4')}

    # For the moment, we assume that ground motions are observed. Later, hazard nodes will be added.
    data['GM_obs'] = {'e1': 30.0,
              'e2': 20.0,
              'e3': 10.0,
              'e4': 2.0,
              'e5': 0.9,
              'e6': 0.6}

    return data


@pytest.fixture(scope='package')
def expected_probs():

    probs = {}
    probs['disconn'] = np.array([0.0096, 0.0011, 0.2102, 0.2102])
    probs['delay'] = np.array([0.0583, 0.0052, 0.4795, 0.4382])
    probs['damage'] = np.array([0.1610,  1,  1,  0.0002,   0,  0.4382])

    return probs


@pytest.fixture(scope='package')
def setup_bridge(data_bridge):
    """
    system with 6 edges, 5 nodes
    (see graph.png)
    """
    arcs = data_bridge['arcs']
    node_coords = data_bridge['node_coords']
    arcs_avg_kmh = data_bridge['arcs_avg_kmh']
    arcs_type = data_bridge['arcs_type']
    GM_obs = data_bridge['GM_obs']
    frag = data_bridge['frag']
    var_ODs = data_bridge['var_ODs']

    # Arcs' states index compatible with variable B index, and C
    #arc_surv = 0
    #arc_fail = 1
    #arc_either = 2

    arc_lens_km = trans.get_arcs_length(arcs, node_coords)

    #arcTimes_h = arcLens_km ./ arcs_Vavg_kmh
    arc_times_h = {k: v/arcs_avg_kmh[k] for k, v in arc_lens_km.items()}

    # create a graph
    G = nx.Graph()
    for k, x in arcs.items():
        G.add_edge(x[0], x[1], weight=arc_times_h[k], label=k)

    for k, v in node_coords.items():
        G.add_node(k, pos=v, label=k)

    # plot graph
    pos = nx.get_node_attributes(G, 'pos')
    edge_labels = nx.get_edge_attributes(G, 'label')

    fig = plt.figure()
    ax = fig.add_subplot()
    nx.draw(G, pos, with_labels=True, ax=ax)
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, ax=ax)
    fig.savefig(HOME.joinpath('graph_bridge.png'), dpi=100)

    # Arcs (components): P(X_i | GM = GM_ob ), i = 1 .. N (= nArc)
    cpms_arc = {}
    vars_arc = {}

    #path_time = trans.get_all_paths_and_times(var_ODs.values(), G, key='time')

    # number of component states: 2 ('surv' or 'fail')
    for k in arcs.keys():
        #vars_arc[k] = variable.Variable(name=str(k), B=[{0}, {1}, {0, 1}], values=['Fail', 'Surv'])
        vars_arc[k] = variable.Variable(name=str(k), values=['Fail', 'Surv'])

        _type = arcs_type[k]
        prob = lognorm.cdf(GM_obs[k], frag[_type]['std'], scale=frag[_type]['med'])
        C = np.array([[0, 1]]).T
        p = np.array([prob, 1-prob])
        cpms_arc[k] = cpm.Cpm(variables = [vars_arc[k]],
                              no_child = 1,
                              C = C,
                              p = p)

    # Travel times (systems): P(OD_j | X1, ... Xn) j = 1 ... nOD
    # e.g., for 'od1': 'e2': 0.0901, 'e3'-'e1': 0.2401
    vars_arc['od1'] = variable.Variable(name='od1',
            values=[np.inf, 0.2401, 0.0901])

    vars_arc['od2'] = variable.Variable(name='od2',
            values=[np.inf, 0.2401, 0.0901])

    vars_arc['od3'] = variable.Variable(name='od3',
            values=[np.inf, 0.1761, 0.0943])

    vars_arc['od4'] = variable.Variable(name='od4',
            values=[np.inf, 0.1997, 0.0707])

    _variables = [vars_arc[k] for k in ['od1', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c7 = np.array([[0,2,0,0,2,2,2],
                   [0,0,0,1,2,2,2],
                   [1,1,0,1,2,2,2],
                   [2,2,1,2,2,2,2]])
    cpms_arc['od1'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c7,
                           p = [1, 1, 1, 1],
                           )

    _variables = [vars_arc[k] for k in ['od2', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c8 = np.array([[0,2,0,0,2,2,2],
                  [0,0,1,0,2,2,2],
                  [1,1,1,0,2,2,2],
                  [2,2,2,1,2,2,2]])
    cpms_arc['od2'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c8,
                           p = [1, 1, 1, 1],
                           )

    _variables = [vars_arc[k] for k in ['od3', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c9 = np.array([[0,2,2,2,2,0,0],
                   [0,2,2,2,0,0,1],
                   [1,2,2,2,1,0,1],
                   [2,2,2,2,2,1,2]])
    cpms_arc['od3'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c9,
                           p = [1, 1, 1, 1],
                           )

    _variables = [vars_arc[k] for k in ['od4', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c10 = np.array([[0,2,2,2,2,0,0],
                   [0,2,2,2,0,1,0],
                   [1,2,2,2,1,1,0],
                   [2,2,2,2,2,2,1]])
    cpms_arc['od4'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c10,
                           p = [1, 1, 1, 1],
                           )
    #probs = {k: cpms_arc[k].p[:, 0].tolist() for k in arcs.keys()}

    return cpms_arc, vars_arc, arcs, var_ODs


@pytest.fixture(scope='package')
def setup_bridge_alt(data_bridge):
    """
    Note the difference of arc's state from the setup_bridge
    """

    arcs = data_bridge['arcs']
    node_coords = data_bridge['node_coords']
    arcs_avg_kmh = data_bridge['arcs_avg_kmh']
    arcs_type = data_bridge['arcs_type']
    GM_obs = data_bridge['GM_obs']
    frag = data_bridge['frag']
    var_ODs = data_bridge['var_ODs']

    # Arcs' states index compatible with variable B index, and C
    #arc_fail = 0
    #arc_surv = 1
    #arc_either = 2

    arc_lens_km = trans.get_arcs_length(arcs, node_coords)

    #arcTimes_h = arcLens_km ./ arcs_Vavg_kmh
    arc_times_h = {k: v/arcs_avg_kmh[k] for k, v in arc_lens_km.items()}

    # create a graph
    G = nx.Graph()
    for k, x in arcs.items():
        G.add_edge(x[0], x[1], time=arc_times_h[k], label=k)

    for k, v in node_coords.items():
        G.add_node(k, pos=v)

    #path_time = trans.get_all_paths_and_times(var_ODs.values(), G, key='time')
    """
    # plot graph
    pos = nx.get_node_attributes(G, 'pos')
    edge_labels = nx.get_edge_attributes(G, 'label')

    fig = plt.figure()
    ax = fig.add_subplot()
    nx.draw(G, pos, with_labels=True, ax=ax)
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, ax=ax)
    fig.savefig(HOME.joinpath('graph.png'), dpi=200)
    """
    # Arcs (components): P(X_i | GM = GM_ob ), i = 1 .. N (= nArc)
    cpms_arc = {}
    vars_arc = {}

    for k in arcs.keys():
        #vars_arc[k] = variable.Variable(name=str(k), B=[{0}, {1}, {0, 1}], values=['Fail', 'Surv'])
        vars_arc[k] = variable.Variable(name=str(k), values=['Fail', 'Surv'])

        _type = arcs_type[k]
        prob = lognorm.cdf(GM_obs[k], frag[_type]['std'], scale=frag[_type]['med'])
        C = np.array([[0, 1]]).T
        p = np.array([prob, 1-prob])
        cpms_arc[k] = cpm.Cpm(variables = [vars_arc[k]],
                              no_child = 1,
                              C = C,
                              p = p)

    # Travel times (systems): P(OD_j | X1, ... Xn) j = 1 ... nOD
    # e.g., for 'od1': 'e2': 0.0901, 'e3'-'e1': 0.2401
    vars_arc['od1'] = variable.Variable(name='od1',
            values=[0.0901, 0.2401, np.inf])

    vars_arc['od2'] = variable.Variable(name='od2',
            values=[0.0901, 0.2401, np.inf])

    vars_arc['od3'] = variable.Variable(name='od3',
            values=[0.0943, 0.1761, np.inf])

    vars_arc['od4'] = variable.Variable(name='od4',
            values=[0.0707, 0.1997, np.inf])

    _variables = [vars_arc[k] for k in ['od1', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c7 = np.array([
    [1,3,2,3,3,3,3],
    [2,2,1,2,3,3,3],
    [3,2,1,1,3,3,3],
    [3,1,1,3,3,3,3]]) - 1

    cpms_arc['od1'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c7,
                           p = [1, 1, 1, 1],
                           )

    _variables = [vars_arc[k] for k in ['od2', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c8 = np.array([
    [1,3,3,2,3,3,3],
    [2,2,2,1,3,3,3],
    [3,2,1,1,3,3,3],
    [3,1,3,1,3,3,3]]) - 1
    cpms_arc['od2'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c8,
                           p = [1, 1, 1, 1],
                           )

    _variables = [vars_arc[k] for k in ['od3', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c9 = np.array([
    [1,3,3,3,3,2,3],
    [2,3,3,3,2,1,2],
    [3,3,3,3,2,1,1],
    [3,3,3,3,1,1,3]]) - 1
    cpms_arc['od3'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c9,
                           p = [1, 1, 1, 1],
                           )

    _variables = [vars_arc[k] for k in ['od4', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    c10 = np.array([
    [1,3,3,3,3,3,2],
    [2,3,3,3,2,2,1],
    [3,3,3,3,2,1,1],
    [3,3,3,3,1,3,1]]) - 1
    cpms_arc['od4'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c10,
                           p = [1, 1, 1, 1],
                           )

    return cpms_arc, vars_arc, arcs, var_ODs


@pytest.fixture
def setup_multi_dest():
    # see graph.png 

    arcs = {'e1': ['n1', 'n2'],
            'e2': ['n1', 'n5'],
            'e3': ['n2', 'n5'],
            'e4': ['n3', 'n4'],
            'e5': ['n3', 'n5'],
            'e6': ['n4', 'n5']}

    arc_times_h = {'e1': [np.inf, 0.15], 'e2': [np.inf, 0.0901], 'e3': [np.inf, 0.0901], 'e4': [np.inf, 0.1054],
                   'e5': [np.inf, 0.0943], 'e6': [np.inf, 0.0707]}

    # create a graph
    G = nx.Graph()
    for k, x in arcs.items():
        G.add_edge(x[0], x[1], weight=arc_times_h[k][1], label=k)

    [G.add_node(f'n{k}') for k in range(1, 6)]

    varis = {k: variable.Variable(name=k, values=v) for k, v in arc_times_h.items()}

    origin = 'n1'
    dests = ['n3', 'n4']

    return G, origin, dests, arcs, varis


def test_prob_delay1(setup_bridge, expected_probs):

    ## Inference - by variable elimination (would not work for large-scale systems)
    # Probability of delay and disconnection
    # Becomes P(OD_1, ..., OD_M) since X_1, ..., X_N are eliminated
    #cpms_arc_cp = cpms_arc.values()

    d_cpms_arc, d_vars_arc, arcs, var_ODs = setup_bridge
    cpms_arc = copy.deepcopy(d_cpms_arc)
    vars_arc = copy.deepcopy(d_vars_arc)

    cpms_arc_cp = list(cpms_arc.values())
    nODs = len(var_ODs)

    # prod_cpms
    # get different variables order
    for i in arcs.keys():

        is_inscope = inference.isinscope([vars_arc[i]], cpms_arc_cp)
        cpm_sel = [y for x, y in zip(is_inscope, cpms_arc_cp) if x]
        cpm_mult = cpm.product(cpm_sel)
        cpm_mult = cpm_mult.sum([vars_arc[i]])

        cpms_arc_cp = [y for x, y in zip(is_inscope, cpms_arc_cp) if x == False]
        cpms_arc_cp.insert(0, cpm_mult)

    # cpms_arc_cp[0].variables = [10, 9, 7, 8]
    #print([x.name for x in cpms_arc_cp[0].variables])

    # Retrieve example results
    # P( OD_j = 3 ), j = 1, ..., M, where State 3 indicates disconnection
    ODs_prob_disconn = np.zeros(nODs)
    # P( (OD_j = 2) U (OD_j = 2) ), j = 1, ..., M, where State 2 indicates the use of the second shortest path (or equivalently, P( (OD_j = 1)^c ), where State 1 indicates the use of the shortest path)
    ODs_prob_delay = np.zeros(nODs)

    #disconn_state = 3 - 1
    disconn_state = 0

    for j, idx in enumerate(var_ODs):

        # Prob. of disconnection
        [cpm_ve] = inference.condition(cpms_arc_cp,
                                       cnd_vars=[vars_arc[idx]],
                                       cnd_states=[disconn_state])
        ODs_prob_disconn[j] = cpm_ve.p.sum(axis=0).item()

        # Prob. of delay
        var_loc = cpms_arc_cp[0].variables.index(vars_arc[idx])
        # except the shortest path (which is state 0)
        rows_to_keep = np.where(cpms_arc_cp[0].C[:, var_loc] < 2)[0]
        cpm_ve = cpms_arc_cp[0].get_subset(rows_to_keep)
        ODs_prob_delay[j] = cpm_ve.p.sum(axis=0).item()

        # Prob. of disconnection alternative
        rows_to_keep = np.where(cpms_arc_cp[0].C[:, var_loc] == disconn_state)[0]
        assert cpms_arc_cp[0].get_subset(rows_to_keep).p.sum() == ODs_prob_disconn[j]

    #plot_delay(ODs_prob_delay, ODs_prob_disconn, var_ODs)

    # Check if the results are the same
    np.testing.assert_array_almost_equal(ODs_prob_disconn, expected_probs['disconn'], decimal=4)
    np.testing.assert_array_almost_equal(ODs_prob_delay, expected_probs['delay'], decimal=4)


def test_prob_delay2(setup_bridge, expected_probs):

    d_cpms_arc, d_vars_arc, arcs, var_ODs = setup_bridge

    cpms_arc = copy.deepcopy(d_cpms_arc)
    vars_arc = copy.deepcopy(d_vars_arc)

    nODs = len(var_ODs)

    ## Repeat inferences again using new functions -- the results must be the same.
    # Probability of delay and disconnection
    M = [cpms_arc[k] for k in list(arcs.keys()) + list(var_ODs.keys())]
    var_elim_order = [vars_arc[i] for i in arcs.keys()]
    M_VE2 = inference.variable_elim(M, var_elim_order)

    # "M_VE2" same as "M_VE"
    # Retrieve example results
    ODs_prob_disconn2 = np.zeros(nODs)
    ODs_prob_delay2 = np.zeros(nODs)

    disconn_state = 0
    for j, idx in enumerate(var_ODs):

        # Prob. of disconnection
        # FIXME 2 -> 3?? 
        #disconn_state = vars_arc[idx].values.index(np.inf) + 1
        # the state of disconnection is assigned an arbitrarily large number 100
        ODs_prob_disconn2[j] = M_VE2.get_prob([vars_arc[idx]], [0])

        # Prob. of delay
        #ODs_prob_delay2[j] = cpm.get_prob(M_VE2, [vars_arc[idx]], [1-1], flag=False) # Any state greater than 1 means delay.
        ODs_prob_delay2[j] = M_VE2.get_prob([vars_arc[idx]], [0]) + M_VE2.get_prob([vars_arc[idx]], [1]) # 0, 1

    # Check if the results are the same
    np.testing.assert_array_almost_equal(ODs_prob_disconn2, expected_probs['disconn'], decimal=4)
    np.testing.assert_array_almost_equal(ODs_prob_delay2, expected_probs['delay'], decimal=4)


def test_prob_delay3(setup_bridge_alt, expected_probs):
    """ same as delay2 but only using one OD"""
    cpms_arc, vars_arc, arcs, var_ODs = setup_bridge_alt
    nODs = len(var_ODs)

    ## Repeat inferences again using new functions -- the results must be the same.
    # Probability of delay and disconnection
    #M = [cpms_arc[k] for k in list(arcs.keys()) + list(var_ODs.keys())]
    M = [cpms_arc[k] for k in list(arcs.keys()) + ['od1']]
    var_elim_order = [vars_arc[i] for i in arcs.keys()]
    M_VE2 = inference.variable_elim(M, var_elim_order)

    # "M_VE2" same as "M_VE"
    # Retrieve example results
    ODs_prob_disconn2 = np.zeros(nODs)
    ODs_prob_delay2 = np.zeros(nODs)

    #disconn_state = 2
    for j, idx in enumerate(['od1']):

        # Prob. of disconnection
        # FIXME 2 -> 3?? 
        #disconn_state = vars_arc[idx].values.index(np.inf) + 1
        # the state of disconnection is assigned an arbitrarily large number 100
        ODs_prob_disconn2[j] = M_VE2.get_prob([vars_arc[idx]], [2])

        # Prob. of delay
        ODs_prob_delay2[j] = M_VE2.get_prob([vars_arc[idx]], [1-1], flag=False) # Any state greater than 1 means delay.

    # Check if the results are the same
    np.testing.assert_array_almost_equal(ODs_prob_disconn2[0], expected_probs['disconn'][0], decimal=4)
    np.testing.assert_array_almost_equal(ODs_prob_delay2[0], expected_probs['delay'][0], decimal=4)


def test_prob_damage(setup_bridge, expected_probs):

    #arc_fail = 1
    #arc_surv = 0

    d_cpms_arc, d_vars_arc, arcs, var_ODs = setup_bridge
    cpms_arc = copy.deepcopy(d_cpms_arc)
    vars_arc = copy.deepcopy(d_vars_arc)

    # City 1 (od1) and 2 (od2) experienced a disruption in getting resources, City 3 (od3) was okay and 4 (od4) is unknown. Probability of damage of roads?
    # A composite state needs be created for City 1 and City 2
    # cf. 
    #B_ = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    #vars_arc['OD1'] = variable.Variable(name='OD1', B=B_,
    #        values=[0.0901, 0.2401, np.inf])
    for idx in var_ODs.keys():
        vars_arc[idx].B.append({1, 2})

    # FIXME: sorting of variables in product
    # # Add observation nodes P( O_j | OD_j ), j = 1, ..., M
    var_ODs_obs = [f'od_obs{i}' for i, _ in enumerate(var_ODs.keys(), 1)]
    # column 0 for od_obs: No disruption or Disruption
    # column 1 for od: 2 (shortest), 1 (alternative), 0 (except the shortest path) 
    C = np.array([[0, 2], [1, 0], [1, 1]])
    for j, idx in zip(var_ODs_obs, var_ODs.keys()):

        vars_arc[j] = variable.Variable(name=j,
            #B=[{0}, {1}],
            values=['No disruption', 'Disruption'])

        _variables = [vars_arc[k] for k in [j, idx]]
        cpms_arc[j]= cpm.Cpm(variables=_variables,
                             no_child=1,
                             C= C,
                             p= [1, 1, 1])

    # using string
    cpm_ve = inference.condition(cpms_arc,
                           cnd_vars=['od_obs1', 'od_obs2', 'od_obs3'],
                           cnd_states=[1, 1, 0])

    """
    assert len(cpm_ve) == 14
    assert [x.name for x in cpm_ve[10].variables] == ['od_obs1', 'od1']
    np.testing.assert_array_equal(cpm_ve[10].C, np.array([[1, 1], [1, 2]]))
    np.testing.assert_array_almost_equal(cpm_ve[10].p, np.array([[1, 1]]).T)

    assert [x.name for x in cpm_ve[11].variables] == ['od_obs2', 'od2']
    np.testing.assert_array_equal(cpm_ve[11].C, np.array([[1, 1], [1, 2]]))
    np.testing.assert_array_almost_equal(cpm_ve[11].p, np.array([[1, 1]]).T)

    assert [x.name for x in cpm_ve[12].variables] == ['od_obs3', 'od3']
    np.testing.assert_array_equal(cpm_ve[12].C, np.array([[0, 0]]))
    np.testing.assert_array_almost_equal(cpm_ve[12].p, np.array([[1]]).T)

    assert [x.name for x in cpm_ve[13].variables] == ['od_obs4', 'od4']
    np.testing.assert_array_equal(cpm_ve[13].C, np.array([[0, 0], [1, 1], [1, 2]]))
    np.testing.assert_array_almost_equal(cpm_ve[13].p, np.array([[1, 1, 1]]).T)
    """

    Mcond_mult = cpm.product(cpm_ve)

    assert Mcond_mult.C.shape == (8, 14)
    #assert [x.name for x in Mcond_mult.variables] == ['e1', 'e2', 'e3', 'e4', 'e5', 'e6', 'od1', 'od2', 'od3', 'od4', 'od_obs1', 'od_obs2', 'od_obs3', 'od_obs4']

    expected_p = np.array([[7.48743285822800e-05,
1.43718130993790e-05,
1.44320505816953e-08,
2.77017153313071e-09,
5.84121979269559e-05,
1.12119762170237e-05,
1.12589696766821e-08,
2.16111197186942e-09]]).T

    #np.testing.assert_array_almost_equal(Mcond_mult.p, expected_p)

    # using string
    Mcond_mult_sum = Mcond_mult.sum(list(var_ODs.keys()) + var_ODs_obs)

    assert Mcond_mult_sum.C.shape == (8, 6)
    #assert [x.name for x in Mcond_mult_sum.variables] == ['e1', 'e2', 'e3', 'e4', 'e5', 'e6']

    expected_p = np.array([[7.48743285822800e-05,
1.43718130993790e-05,
1.44320505816953e-08,
2.77017153313071e-09,
5.84121979269559e-05,
1.12119762170237e-05,
1.12589696766821e-08,
2.16111197186942e-09]]).T

    #np.testing.assert_array_almost_equal(Mcond_mult_sum.p, expected_p)

    # P( X_i = 2 | OD_1 = 2, OD_2 = 2, OD_3 = 1 ), i = 1, ..., N
    arc_fail = 0
    arcs_prob_damage = np.zeros(len(arcs))
    for j, i in enumerate(arcs.keys()):

        iM = Mcond_mult_sum.sum([vars_arc[i]], 0)
        iM_fail = iM.condition(cnd_vars=[vars_arc[i]],
                               cnd_states=[arc_fail])
        fail_prob = iM_fail.p / iM.p.sum(axis=0)
        if fail_prob.size:
            arcs_prob_damage[j] = fail_prob.item()

    # Check if the results are the same
    np.testing.assert_array_almost_equal(arcs_prob_damage,  expected_probs['damage'], decimal=4)

    """
    plt.figure()

    barWidth = 0.25
    r1 = np.arange(len(arcs))
    r2 = [x + barWidth for x in r1]

    plt.bar(r1, 1-arcs_prob_damage, color='#7f6d5f', width=barWidth, edgecolor='white', label='Survival')
    plt.bar(r2, arcs_prob_damage, color='#557f2d', width=barWidth, edgecolor='white', label='Failure')

    plt.xlabel( 'Arc' )
    plt.ylabel( 'Probability' )
    plt.legend()
    plt.savefig(HOME.joinpath('figure2s.png'), dpi=200)
    plt.close()
    """

def plot_delay(ODs_prob_delay, ODs_prob_disconn, var_ODs):

    # plot
    plt.figure()
    nODs = len(var_ODs)
    # set width of bars
    barWidth = 0.25
    # Set position of bar on X axis
    r1 = np.arange(nODs)
    r2 = [x + barWidth for x in r1]

    # Make the plot
    plt.bar(r1, ODs_prob_disconn, color='#7f6d5f', width=barWidth, edgecolor='white', label='Disconnection')
    plt.bar(r2, ODs_prob_delay, color='#557f2d', width=barWidth, edgecolor='white', label='Delay')

    # Add xticks on the middle of the group bars
    plt.xlabel('OD pair', fontweight='bold')
    plt.xticks([r + 0.5*barWidth for r in range(nODs)], [f'{x}' for x in var_ODs.keys()])

    # Create legend & Show graphic
    plt.legend()
    plt.savefig(HOME.joinpath('figure1s.png'), dpi=200)
    plt.close()


def test_get_arcs_length():

    node_coords = {1: [-2, 3],
                   2: [-2, -3],
                   3: [2, -2],
                   4: [1, 1],
                   5: [0, 0]}

    arcs = {1: [1, 2],
            2: [1,5],
            3: [2,5],
            4: [3,4],
            5: [3,5],
            6: [4,5]}

    result = trans.get_arcs_length(arcs, node_coords)

    expected = {1: 6.0,
                2: 3.6056,
                3: 3.6056,
                4: 3.1623,
                5: 2.8284,
                6: 1.4142}

    pd.testing.assert_series_equal(pd.Series(result), pd.Series(expected), rtol=1.0e-3)


def test_get_all_paths_and_times():

    arcs = {1: [1, 2],
            2: [1, 5],
            3: [2, 5],
            4: [3, 4],
            5: [3, 5],
            6: [4, 5]}

    arc_times_h = {1: 0.15, 2: 0.0901, 3: 0.0901, 4: 0.1054,
                   5: 0.0943, 6: 0.0707}

    G = nx.Graph()
    for k, x in arcs.items():
        G.add_edge(x[0], x[1], weight=arc_times_h[k], label=k, key=k)

    ODs = [(5, 1), (5, 2), (5, 3), (5, 4)]

    path_time = trans.get_all_paths_and_times(ODs, G)

    expected = {(5, 1): [([2], 0.0901),
                         ([3, 1], 0.2401)],
                (5, 2): [([2, 1], 0.2401),
                         ([3], 0.0901)],
                (5, 3): [([5], 0.0943),
                         ([6, 4], 0.1761)],
                (5, 4): [([5, 4], 0.1997),
                         ([6], 0.0707)],
                }


def test_get_path_time_idx1():

    path_time =[(['e2'], 0.0901), (['e3', 'e1'], 0.24009999999999998)]

    vari = variable.Variable(name='od1', values=[np.inf, 0.2401, 0.0901])

    result = trans.get_path_time_idx(path_time, vari)

    expected = [([], np.inf, 0), (['e2'], 0.0901, 2), (['e3', 'e1'], 0.24009999999999998, 1)]

    assert result==expected

    path_time =[(['e3', 'e1'], 0.24009999999999998), (['e2'], 0.0901)]

    result = trans.get_path_time_idx(path_time, vari)

    assert result==expected


def test_get_path_time_idx2():

    path_time =[(['e2'], 0.0901), (['e3', 'e1'], 0.24009999999999998)]

    vari = variable.Variable(name='od1', values=[np.inf, 0.2401, 0.0901])

    result = trans.get_path_time_idx(path_time, vari)

    expected = [([], np.inf, 0), (['e2'], 0.0901, 2), (['e3', 'e1'], 0.24009999999999998, 1)]

    assert result==expected


def test_eval_sys_state():

    arc_surv = 1
    arc_fail = 0

    arcs = {'e1': ['n1', 'n2'],
            'e2': ['n1', 'n5'],
            'e3': ['n2', 'n5'],
            'e4': ['n3', 'n4'],
            'e5': ['n3', 'n5'],
            'e6': ['n4', 'n5']}

    arc_times_h = {'e1': 0.15, 'e2': 0.0901, 'e3': 0.0901, 'e4': 0.1054,
                   'e5': 0.0943, 'e6': 0.0707}

    G = nx.Graph()
    for k, x in arcs.items():
        G.add_edge(x[0], x[1], weight=arc_times_h[k], label=k)

    ODs = [('n5', 'n1'), ('n5', 'n2'), ('n5', 'n3'), ('n5', 'n4')]

    arcs_state = {'e1': arc_surv,
                  'e2': arc_surv,
                  'e3': arc_surv,
                  'e4': arc_surv,
                  'e5': arc_surv,
                  'e6': arc_surv}

    vars_od1 = variable.Variable(name='od1',
            values=[np.inf, 0.2401, 0.0901])

    path_time = trans.get_all_paths_and_times([ODs[0]], G, 'weight')[ODs[0]]
    path_time.append(([], np.inf))

    # refering variable
    path_time_idx = []
    for x in path_time:
        idx = [i for i, y in enumerate(vars_od1.values) if isclose(x[1], y)]
        try:
            path_time_idx.append((*x, idx[0]))
        except IndexError:
            print('path_time incompatible with variable')

    # sort by increasing number of edges
    path_time_idx = sorted(path_time_idx, key=lambda x: len(x[0]))

    # The shortest route (2) available
    sys_state = trans.eval_sys_state(path_time_idx, arcs_state, arc_surv)
    assert sys_state == 2

    arcs_state = {'e1': arc_surv,
                  'e2': arc_fail,
                  'e3': arc_surv,
                  'e4': arc_surv,
                  'e5': arc_surv,
                  'e6': arc_surv}

    # The second shortest route (1,3) available
    sys_state = trans.eval_sys_state(path_time_idx, arcs_state, arc_surv)
    assert sys_state == 1

    arcs_state = {'e1': arc_surv,
                  'e2': arc_fail,
                  'e3': arc_fail,
                  'e4': arc_surv,
                  'e5': arc_surv,
                  'e6': arc_surv}

    # No path available
    sys_state = trans.eval_sys_state(path_time_idx, arcs_state, arc_surv)
    assert sys_state == 0


def test_get_connectivity_given_comps1(setup_sys_rbd):

    varis, arcs, G = setup_sys_rbd

    comps_st = {f'x{i}': 1 for i in range(1, 9)}
    #comps_st.update({'sink': 1, 'source': 1})
    od_pair = ('source', 'sink')

    #_varis = {k: varis[k] for k in comps_st.keys()}
    path_edges, path_nodes = trans.get_connectivity_given_comps(comps_st,  G, od_pair)
    assert path_nodes == ['source', 'x1', 'x7', 'x8', 'sink']
    assert path_edges == ['e1', 'e10', 'e8', 'e9']

    # failure
    _comps_st = comps_st.copy()
    _comps_st['x7'] = 0
    path_edges, path_nodes = trans.get_connectivity_given_comps(_comps_st,  G, od_pair)
    assert path_edges == []

    # success
    _comps_st = comps_st.copy()
    _comps_st.update({'x2': 0, 'x3': 0, 'x1': 0})
    path_edges, path_nodes = trans.get_connectivity_given_comps(_comps_st, G, od_pair)
    assert path_nodes == ['source', 'x4', 'x5', 'x6', 'x7', 'x8', 'sink']
    assert path_edges == ['e4', 'e5', 'e6', 'e7', 'e8', 'e9']

    # failure
    _comps_st = comps_st.copy()
    _comps_st.update({'x4': 0, 'x2': 0, 'x1': 0, 'x3': 0})
    path_edges, path_nodes = trans.get_connectivity_given_comps(_comps_st, G, od_pair)
    assert path_edges == []

    # success
    _comps_st = comps_st.copy()
    _comps_st.update({'x1': 0, 'x2': 0, 'x3': 0})
    od_pair = ('x4', 'x6')
    path_edges, path_nodes = trans.get_connectivity_given_comps(_comps_st, G, od_pair)
    assert path_nodes == ['x4', 'x5', 'x6']
    assert path_edges == ['e5', 'e6']


def test_get_connectivity_given_comps2(setup_sys_rbd):

    varis, arcs, G = setup_sys_rbd

    comps_st = {f'e{i}': 1 for i in range(1, 13)}
    od_pair = ('source', 'sink')
    #_varis = {k: varis[k] for k in comps_st.keys()}
    path_e, path_n = trans.get_connectivity_given_comps(comps_st,  G, od_pair)
    assert path_n == ['source', 'x1', 'x7', 'x8', 'sink']
    assert path_e == ['e1', 'e10', 'e8', 'e9']

    # failure
    _comps_st = comps_st.copy()
    _comps_st['e8'] = 0
    path_e, path_n = trans.get_connectivity_given_comps(_comps_st,  G, od_pair)
    assert path_n == []

    # failure from node and edge failure
    _comps_st = comps_st.copy()
    _comps_st['x7'] = 0
    #_varis.update({'x7': varis['x7']})
    path_e, path_n = trans.get_connectivity_given_comps(_comps_st,  G, od_pair)
    assert path_n == []


def test_get_connectivity_given_comps4():

    x_star = {'e1': 1, 'e2': 0, 'e3': 1}

    varis = {}
    for i in range(1, 4):
        varis[f'e{i}'] = variable.Variable(name=f'e{i}', values=['Fail', 'Survive'])

    od_pair = ('n1', 'n3')
    edges = {
         'e1': {'origin': 'n1', 'destination': 'n2', 'link_capacity': None, 'weight': 1.0},
         'e2': {'origin': 'n2', 'destination': 'n3', 'link_capacity': None, 'weight': 1.0},
         'e3': {'origin': 'n2', 'destination': 'n3', 'link_capacity': None, 'weight': 1.0},
         }

    G = nx.MultiDiGraph()
    G.add_edge('n1', 'n2', label='e1', key='e1', weight=1)
    G.add_edge('n2', 'n3', label='e2', key='e2', weight=1)
    G.add_edge('n2', 'n3', label='e3', key='e3', weight=1)
    G.add_node('n1', key='n1', label='n1')
    G.add_node('n2', key='n2', label='n2')
    G.add_node('n3', key='n3', label='n3')

    s_path_edges, s_path_nodes = trans.get_connectivity_given_comps(x_star, G, od_pair)

    assert s_path_edges == ['e1', 'e3']
    assert s_path_nodes == ['n1', 'n2', 'n3']

    #sys_fun = trans.sys_fun_wrap(od_pair, edges, varis)

    #sys_val, sys_st, comp_st_min = sys_fun(x_star)

    #assert sys_st == 's'
    #assert sys_val == None
    #assert comp_st_min == {'e1': 1, 'e3': 1}


def test_sf_min_path(setup_sys_rbd):

    varis, arcs, G = setup_sys_rbd

    # edges
    comps_st = {f'e{i}': 1 for i in range(1, 13)}
    od_pair = ('source', 'sink')

    d_time, sys_st, min_comps_st = trans.sf_min_path(comps_st, G, od_pair)
    assert sys_st == 's'
    assert min_comps_st == {'e1': 1, 'e10': 1, 'e8': 1, 'e9':1}

    # nodes
    comps_st = {f'x{i}': 1 for i in range(1, 9)}
    d_time, sys_st, min_comps_st = trans.sf_min_path(comps_st, G, od_pair)
    assert sys_st == 's'
    assert min_comps_st == {'x1': 1, 'x7': 1, 'x8': 1}

    # nodes: failure
    _comps_st = {f'x{i}': 1 for i in range(1, 9)}
    _comps_st['x7'] = 0
    d_time, sys_st, min_comps_st = trans.sf_min_path(_comps_st, G, od_pair)
    assert sys_st == 'f'
    assert min_comps_st == {}


def test_get_time_and_path_multi_dest1(setup_multi_dest):

    G, origin, dests, _, varis = setup_multi_dest

    arcs_state = {'e1': 1,
                  'e2': 0,
                  'e3': 1,
                  'e4': 1,
                  'e5': 1,
                  'e6': 0}

    d_time1, path1, path1e = trans.get_time_and_path_multi_dest(arcs_state, G, origin, dests, varis)

    assert d_time1 == pytest.approx(0.3344, rel=1.0e-4)
    assert path1 == ['n1', 'n2', 'n5', 'n3']
    assert path1e == ['e1', 'e3', 'e5']

    arcs_state['e3'] = 0 # no path available in this case
    d_time2, path2, path2_e = trans.get_time_and_path_multi_dest(arcs_state, G, origin, dests, varis)

    assert d_time2 == np.inf


def test_sys_fun_wrap1(setup_multi_dest):

    G, origin, dests, arcs, varis = setup_multi_dest

    arcs_state = {'e1': 1,
                  'e2': 0,
                  'e3': 1,
                  'e4': 1,
                  'e5': 1,
                  'e6': 0}

    od_pair = {'origin': origin, 'dests': dests}
    sys_fun = trans.sys_fun_wrap(G, od_pair, varis=varis, thres=0.5)
    d_time1, sys_st1, min_comps_st1 = sys_fun(arcs_state)

    assert d_time1 == pytest.approx(0.3344, rel=1.0e-4)
    assert sys_st1 == 's'
    assert min_comps_st1 == {'e1': 1, 'e3': 1, 'e5': 1}

    arcs_state['e3'] = 0
    d_time2, sys_st2, min_comps_st2 = sys_fun(arcs_state)

    assert d_time2 == np.inf
    assert sys_st2 == 'f'
    assert min_comps_st2 == {}


def test_get_edge_from_nodes():

    G = nx.MultiGraph()
    G.add_edge('n1', 'n2', weight=1, label='e1')
    G.add_edge('n2', 'n3', weight=1, label='e2')
    G.add_edge('n2', 'n3', weight=1, label='e3')

    G.add_node('n1', pos=(0, 0), label='n1')
    G.add_node('n2', pos=(1, 0), label='n2')
    G.add_node('n3', pos=(2, 0), label='n3')

    comps_st = {'e1': 1, 'e2': 1, 'e3': 1}
    od_pair = ('n1', 'n3')
    path_nodes = ['n1', 'n2', 'n3']
    path_edge = trans.get_edge_from_nodes(G, path_nodes)

    assert path_edge == ['e1', 'e2']





# replicate demoMCSproduct.ipynb

import numpy as np
import pytest
import copy
import pdb
import networkx as nx
from scipy.stats import lognorm

np.set_printoptions(precision=3)

from mbnpy import cpm, variable, trans, inference


@pytest.fixture(scope='package')
def setup_sys(data_bridge):

    cpms = {}
    varis = {}

    low = 0
    high = 1
    #p_low = 0.95
    #p_high = 0.05

    # hazard
    varis['haz'] = variable.Variable(name='haz', values=['low', 'high'])

    C = np.array([[0, 1]]).T
    p = np.array([0.95, 0.05])
    cpms['haz'] = cpm.Cpm( variables = [varis['haz']], no_child = 1, C = C, p = p )

    # 
    node_coords = data_bridge['node_coords']

    var_ODs = data_bridge['var_ODs']

    arcs = data_bridge['arcs']

    frag = data_bridge['frag']

    arcs_type = data_bridge['arcs_type']

    arcs_avg_kmh = data_bridge['arcs_avg_kmh']

    # For the moment, we assume that ground motions are observed. Later, hazard nodes will be added.
    GM_obs = data_bridge['GM_obs']

    arc_lens_km = trans.get_arcs_length(arcs, node_coords)

    #arcTimes_h = arcLens_km ./ arcs_Vavg_kmh
    arc_times_h = {k: v/arcs_avg_kmh[k] for k, v in arc_lens_km.items()}

    # create a graph
    G = nx.Graph()
    for k, x in arcs.items():
        G.add_edge(x[0], x[1], time=arc_times_h[k], label=k)

    for k, v in node_coords.items():
        G.add_node(k, pos=v)

    path_time = trans.get_all_paths_and_times(var_ODs.values(), G, key='time')

    # Arcs (components): P(X_i | GM = GM_ob ), i = 1 .. N (= nArc)
    # Arcs' states index compatible with variable B index, and C
    arc_surv = 0
    arc_fail = 1
    arc_either = 2

    C = np.array([[arc_surv, low], [arc_fail, low],
              [arc_surv, high], [arc_fail, high]])

    for k in arcs.keys():
        varis[k] = variable.Variable(name=k, values=['Surv', 'Fail'])

        _type = arcs_type[k]
        prob = [lognorm.cdf(GM_obs[k], frag[_type]['std'], scale=frag[_type]['med']),
                lognorm.cdf(GM_obs[k]*1.5, frag[_type]['std'], scale=frag[_type]['med'])]

        p = np.array([1-prob[0], prob[0], 1-prob[1], prob[1]])
        cpms[k] = cpm.Cpm(variables = [varis[k], varis['haz']],
                              no_child = 1,
                              C = C,
                              p = p)

    # Travel times (systems): P(OD_j | X1, ... Xn) j = 1 ... nOD
    varis['od1'] = variable.Variable(name='od1',
        values=[0.0901, 0.2401, np.inf])

    _variables = [varis[k] for k in ['od1', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]

    c7 = np.array([
        [1,3,1,3,3,3,3],
        [2,1,2,1,3,3,3],
        [3,1,2,2,3,3,3],
        [3,2,2,3,3,3,3]]) - 1

    cpms['od1'] = cpm.Cpm(variables= _variables,
                           no_child = 1,
                           C = c7,
                           p = [1, 1, 1, 1],
                           )

    return arcs, cpms, varis


def test_exact(setup_sys):

    arcs, cpms, varis = setup_sys

    #cpm_od1 = cpms.variable_elim( cpms, ['haz', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6'] )

    cpms_cp = list(cpms.values())
    for i in ['haz'] + list(arcs.keys()):
        is_inscope = inference.isinscope([varis[i]], cpms_cp)
        cpm_sel = [y for x, y in zip(is_inscope, cpms_cp) if x]
        cpm_mult = cpm.product(cpm_sel)
        cpm_mult = cpm_mult.sum([varis[i]])

        cpms_cp = [y for (x,y) in zip(is_inscope, cpms_cp) if x == False]
        cpms_cp.insert(0, cpm_mult)

    prob_disruption = cpms_cp[0].p[1][0] + cpms_cp[0].p[2][0]
    prob_disconnection = cpms_cp[0].p[2][0]

    assert prob_disruption == pytest.approx(0.0634, abs=1.0e-4)
    assert prob_disconnection == pytest.approx(0.012, abs=1.0e-4)


def test_variable_elim(setup_sys):

    arcs, cpms, varis = setup_sys

    #M = [cpms[k] for k in ['haz'] + list(arcs.keys()) + ['od1']]
    M = [cpms[k] for k in ['od1'] + ['haz'] + list(arcs.keys())]
    elim = [varis[k] for k in ['haz', 'e1', 'e2', 'e3', 'e4', 'e5', 'e6']]
    cpm_od1 = inference.variable_elim(M, elim)

    # prob disruption
    assert cpm_od1.p[1] + cpm_od1.p[2] == pytest.approx(0.0634, abs=1.0e-4)
    # prob disconnection
    assert cpm_od1.p[2] == pytest.approx(0.012, abs=1.0e-4)


#@pytest.mark.skip('FIXME')
def test_mcs_product(setup_sys):

    arcs, cpms, varis = setup_sys

    nsample = 10
    var_mcs = ['haz']
    cpm_h_mcs = inference.mcs_product({'haz': cpms['haz']}, nsample)
    #assert cpm_h_mcs.C.sum() == pytest.approx(nsample * cpms['haz'].p[1], abs=3)

    cpms_cp = list(cpms.values())
    elim_order = ['e1', 'e2', 'e3', 'e4', 'e5', 'e6'] # except 'haz' and 'od1'
    for _c, _p in zip(cpm_h_mcs.Cs, cpm_h_mcs.q):
        cpms_c = inference.condition(cpms, [varis['haz']], _c.tolist() )

        for i in elim_order:
            is_inscope = inference.isinscope([varis[i]], cpms_cp)
            cpm_sel = [y for x, y in zip(is_inscope, cpms_cp) if x]

            if cpm_sel:
                cpm_mult = cpm.product(cpm_sel)
                cpm_mult = cpm_mult.sum([varis[i]])

                cpms_cp = [y for x, y in zip(is_inscope, cpms_cp) if not x]
                cpms_cp.insert(0, cpm_mult)

    cpm_od1 = cpms_cp[0].product(cpms_cp[1])

    # prob disruption
    prob_disruption = cpm_od1.p[(cpm_od1.C[:, 0] == 1) | (cpm_od1.C[:, 0] ==2)].sum()
    assert prob_disruption == pytest.approx(0.0634, abs=1.0e-4)

    # prob disconnection
    prob_disconnection = cpm_od1.p[cpm_od1.C[:, 0] == 2].sum()
    assert prob_disconnection == pytest.approx(0.012, abs=1.0e-4)





import numpy as np
import itertools
import pdb
#from dask.distributed import Variable

from mbnpy import variable, cpm, branch, trans, inference, brc


def setup_model(cfg):
    """
    cfg: instance of config class

    """
    # path_times by od 

    path_times = trans.get_all_paths_and_times(cfg.infra['ODs'].values(), cfg.infra['G'], key='weight')
    branches = get_branches(cfg, path_times)

    # combination of multiple ODs and scenarios
    od_scen_pairs = itertools.product(cfg.infra['ODs'].keys(), cfg.scenarios['scenarios'].keys())

    cpms_by_od_scen = {}
    varis_by_od_scen = {}

    for od, scen in od_scen_pairs:

        cpms_by_od_scen[(od, scen)], varis_by_od_scen[(od, scen)] = model_given_od_scen(cfg, path_times, od, scen, branches[od])

    return cpms_by_od_scen, varis_by_od_scen


def get_branches_by_od(cfg):

    # variables
    varis = {}
    probs = {}
    for k, v in cfg.infra['edges'].items():
        v_scen = cfg.scenarios['scenarios']['s1'][k]
        varis[k] = variable.Variable(name=k, values = v_scen)
        probs[k] = (np.ones_like(v_scen)/len(v_scen)).tolist()


    # Intact state of component vector: zero-based index
    comps_st_itc = {k: len(v.values) - 1 for k, v in varis.items()} # intact state (i.e. the highest state)

    thres = 2
    st_br_to_cs = {'f': 0, 's': 1, 'u': 2}

    csys_by_od, varis_by_od = {}, {}
    # branches by od_pair
    for k, od_pair in cfg.infra['ODs'].items():
        d_time_itc, _, path_itc = trans.get_time_and_path_given_comps(comps_st_itc, cfg.infra['G'], od_pair, varis)

        # system function
        sys_fun = trans.sys_fun_wrap(cfg.infra['G'], od_pair, varis, thres * d_time_itc)
        brs, rules, _, _ = brc.run(probs=probs, sys_fun=sys_fun, max_nb=cfg.max_branches, max_sf=cfg.max_branches)

        csys_by_od[k], varis_by_od[k] = brc.get_csys(brs, varis, st_br_to_cs)

    return csys_by_od, varis_by_od


def get_branches(cfg, path_times):

    # FIXME: only works for binary ATM
    lower = {k: 0 for k, _ in cfg.infra['edges'].items()}
    upper = {k: 1 for k, _ in cfg.infra['edges'].items()}

    # set of branches by od pair
    branches = {}
    for k, v in cfg.infra['ODs'].items():
        values = [np.inf] + sorted([y for _, y in path_times[v]], reverse=True)
        #varis = variable.Variable(name=k, B=[{i} for i in range(len(values))], values=values)
        varis = variable.Variable(name=k, values=values)

        path_time_idx = trans.get_path_time_idx(path_times[v], varis)

        fl = trans.eval_sys_state(path_time_idx, lower, 1)
        fu = trans.eval_sys_state(path_time_idx, upper, 1)

        bstars = [branch.Branch(lower, upper, fl, fu)]

        branch.branch_and_bound(bstars, path_time_idx, arc_cond=1, output_path=cfg.output_path, key=cfg.key)

        branches[k] = branch.get_sb_saved_from_job(cfg.output_path,
                cfg.key)

    return branches


def model_given_od_scen(cfg, path_times, od, scen, branches):

    # Arcs (components): P(X_i | GM = GM_ob ), i = 1 .. N (= nArc)
    cpms = {}
    varis = {}

    # FIXME: only works for binary ATM
    #B = np.vstack([np.eye(cfg.no_ds), np.ones(cfg.no_ds)])

    # scenario dependent
    for k, values in cfg.scenarios['scenarios'][scen].items():
        #B = [{i} for i in range(cfg.no_ds)]
        #B.append({i for i in range(cfg.no_ds)})
        varis[k] = variable.Variable(name=k, values=cfg.scenarios['damage_states'])
        cpms[k] = cpm.Cpm(variables = [varis[k]],
                  no_child = 1,
                  C = np.arange(len(values))[:, np.newaxis],
                  p = values)

    # Travel times (systems): P(OD_j | X1, ... Xn) j = 1 ... nOD
    values = [np.inf] + sorted([y for _, y in path_times[cfg.infra['ODs'][od]]], reverse=True)
    #varis[od] = variable.Variable(name=od, B=[{i} for i in range(len(values))], values=values)
    varis[od] = variable.Variable(name=od, values=values)

    variables = {k: varis[k] for k in cfg.infra['edges'].keys()}
    c = branch.get_cmat(branches, variables)

    cpms[od] = cpm.Cpm(variables = [varis[od]] + list(variables.values()),
                       no_child = 1,
                       C = c,
                       p = np.ones(c.shape[0]),
                       )

    return cpms, varis


def compute_prob(cfg, cpms, varis, var_elim, key, idx_state, flag):
    """
    var_elim: list of variable to be eliminated
    """

    assert isinstance(var_elim, list), 'var_elim should be a list'
    assert isinstance(key, str), 'key should be a str'

    ## Repeat inferences again using new functions -- the results must be the same.
    # Probability of delay and disconnection
    #M = [cpms_arc[k] for k in list(arcs.keys()) + list(var_ODs.keys())]
    M = [cpms[k] for k in var_elim + [key]]
    var_elim = [varis[i] for i in var_elim]
    M_VE2 = inference.variable_elim(M, var_elim)

    # Retrieve example results
    # Prob. of disconnection
    #prob = np.zeros(len(varis[key].values))
    #for idx_state in enumerate(varis[key].values):
    prob = M_VE2.get_prob([varis[key]], [idx_state], flag)

    # Prob. of delay
    #ODs_prob_delay2[j] = cpm.get_prob(M_VE2, [vars_arc[idx]], [1-1], flag=False) # Any state greater than 1 means delay.

    return prob, M_VE2




import numpy as np

from mbnpy.cpm import ismember


def bnb_sys(comp_states, info):
    """
    Parameters
    ----------
    "info": a dict that contains any information about a given problem
        path
        path_time
        arcs
        max_state
    comp_states: list-like

    Output:
    "state": a positve integer indicating a computed system state
    "val": any value that shows what the "state" means if unnecessary, can be left empty
    "result": a structure with any field that is required for the functions "nextComp" and "nextState"
    """
    result = {}

    assert isinstance(info, dict), 'info should be a dict'
    assert isinstance(info['path'], list), 'path should be a list'
    assert isinstance(info['time'], (list, np.ndarray)), 'path_time should be a list-like'
    assert isinstance(info['arcs'], (list, np.ndarray)), 'arcs should be a list-like'

    if isinstance(comp_states, list):
        comp_states = np.array(comp_states)

    if isinstance(info['time'], list):
        info['time'] = np.array(info['time'])

    if isinstance(info['arcs'], np.ndarray):
        info['arcs'] = info['arcs'].tolist()


    path_time = info['time']

    # Ensure shorter paths to be considered first
    path_time = np.sort(path_time)
    path_sort_idx = path_time.argsort().tolist()
    path = [info['path'][i] for i in path_sort_idx]

    # Find the shortest path possible
    idx = np.where(comp_states==info['max_state'])[0]
    try:
        surv_comps = [info['arcs'][i] for i in idx]
    except TypeError:
        surv_comps = []
    is_path_conn = [all(ismember(x, surv_comps)[0]) for x in path]
    is_path_conn = np.where(is_path_conn)[0].tolist()

    # Result
    if is_path_conn:
        is_path_conn = is_path_conn[0] # take the first
        #FIXME
        state = path_sort_idx[is_path_conn]
        val = path_time[is_path_conn]
        result['path'] = path[is_path_conn]
    else:
        #FIXME
        state = len(path_time) # there is no path available
        val = np.inf
        result['path'] = []

    return state, val, result


def bnb_next_comp(cand_comps, down_res, up_res, info):
    """
    Parameters
    ----------
    "info": a dict that contains any information about a given problem
        path
        path_time
        arcs

    comp_states: list-like

    Output:
    "state": a positve integer indicating a computed system state
    "val": any value that shows what the "state" means if unnecessary, can be left empty
    "result": a structure with any field that is required for the functions "nextComp" and "nextState"
    """
    assert isinstance(info['path'], list), 'path should be a list'
    assert isinstance(info['time'], (list, np.ndarray)), 'path_time should be a list-like'
    assert isinstance(info['arcs'], (list, np.ndarray)), 'arcs should be a list-like'

    if isinstance(info['time'], list):
        info['time'] = np.array(info['time'])

    path = info['path']
    path_time = info['time']

    path_sort_idx = path_time.argsort()
    path_time = np.sort(path_time)

    comps_order = []
    _diff = []
    for _path in path:
        # Do not change the order so that components on shorter paths come forward.
        [comps_order.append(x) for x in _path if x not in comps_order]

    [_diff.append(x) for x in info['arcs'] if x not in comps_order]
    [comps_order.append(x) for x in _diff if x not in comps_order]

    next_comp_idx = ismember(cand_comps, comps_order)[0]
    next_comp_idx = np.where(next_comp_idx)[0].tolist()[0]
    next_comp = cand_comps[next_comp_idx]

    return next_comp


def bnb_next_state(next_comp, bound, down_res, up_res, info):
    # In this example, there is only two states, so there's nothing much to do
    next_state = 1

    return next_state



import numpy as np
import pytest
import copy
import pdb
import itertools
from pathlib import Path

from mbnpy import config, model, trans, variable, branch


HOME = Path(__file__).absolute().parent


@pytest.fixture(scope='session')
def setup_road():

    cfg = config.Config(HOME.joinpath('../demos/road/test.json'))
    cpms, varis = model.setup_model(cfg)

    return cpms, varis, cfg

@pytest.mark.skip('json')
def test_setup_model(setup_road, setup_bridge):

    d_cpms, d_varis, cfg = setup_road
    cpms = copy.deepcopy(d_cpms)
    varis = copy.deepcopy(d_varis)

    expected, _, _, _ = setup_bridge

    od_scen_pairs = itertools.product(cfg.infra['ODs'].keys(), cfg.scenarios['scenarios'].keys())

    for od, scen in od_scen_pairs:

        np.testing.assert_array_equal(cpms[(od, scen)][od].C, expected[od].C)


@pytest.mark.skip('json')
def test_compute_prob1(setup_road, expected_probs):

    d_cpms, d_varis, cfg = setup_road
    cpms = copy.deepcopy(d_cpms)
    varis = copy.deepcopy(d_varis)

    var_elim = list(cfg.infra['edges'].keys())

    od_scen_pairs = itertools.product(cfg.infra['ODs'].keys(), cfg.scenarios['scenarios'].keys())

    map_dic = {f'od{i+1}': i for i in range(4)}
    for od, scen in od_scen_pairs:

        prob, _ = model.compute_prob(cfg, cpms[(od, scen)], varis[(od, scen)], var_elim, od, 0, flag=True)
        print(prob, expected_probs['disconn'][map_dic[od]])
        assert expected_probs['disconn'][map_dic[od]] == pytest.approx(prob, abs=0.0001)


@pytest.mark.skip('json')
def test_compute_prob2(setup_road, expected_probs):

    d_cpms, d_varis, cfg = setup_road
    cpms = copy.deepcopy(d_cpms)
    varis = copy.deepcopy(d_varis)

    var_elim = list(cfg.infra['edges'].keys())

    od_scen_pairs = itertools.product(cfg.infra['ODs'].keys(), cfg.scenarios['scenarios'].keys())

    map_dic = {f'od{i+1}': i for i in range(4)}
    for od, scen in od_scen_pairs:

        prob, _ = model.compute_prob(cfg, cpms[(od, scen)], varis[(od, scen)], var_elim, od, 2, flag=False)

        assert expected_probs['delay'][map_dic[od]] == pytest.approx(prob, abs=0.0001)


@pytest.mark.skip('json')
def test_get_branches(setup_bridge):

    expected, _, _, _ = setup_bridge

    cfg = config.Config(HOME.joinpath('../demos/road/test.json'))

    path_times = trans.get_all_paths_and_times(cfg.infra['ODs'].values(), cfg.infra['G'], key='weight')
    pdb.set_trace()
    branches = model.get_branches(cfg, path_times)

    assert len(branches) == len(cfg.infra['ODs'])

    # FIXME: only works for binary ATM
    varis = {}
    #B = np.vstack([np.eye(cfg.no_ds), np.ones(cfg.no_ds)])
    for k in cfg.infra['edges'].keys():
        B = [{i} for i in range(cfg.no_ds)]
        B.append({i for i in range(cfg.no_ds)})
        #varis[k] = variable.Variable(name=k, B=B.copy(), values=cfg.scenarios['damage_states'])
        varis[k] = variable.Variable(name=k, values=cfg.scenarios['damage_states'])

    for od, value in branches.items():
        values = sorted([y for _, y in path_times[cfg.infra['ODs'][od]]]) + [np.inf]
        #varis[od] = variable.Variable(name=od, B=[{i} for i in range(len(values))], values=values)
        varis[od] = variable.Variable(name=od, values=values)

        variables = {k: varis[k] for k in cfg.infra['edges'].keys()}
        c = branch.get_cmat(value, variables)

        np.testing.assert_array_equal(c, expected[od].C)


@pytest.mark.skip('json')
def test_model_given_od_scen(setup_bridge):

    expected, _, _, _ = setup_bridge

    cfg = config.Config(HOME.joinpath('../demos/road/test.json'))

    path_times = trans.get_all_paths_and_times(cfg.infra['ODs'].values(), cfg.infra['G'], key='weight')

    branches = model.get_branches(cfg, path_times)

    od_scen_pairs = itertools.product(cfg.infra['ODs'].keys(), cfg.scenarios['scenarios'].keys())

    for od, scen in od_scen_pairs:

        cpms, varis = model.model_given_od_scen(cfg, path_times, od, scen, branches[od])

        assert len(cpms) == 7

        np.testing.assert_array_equal(cpms[od].C, expected[od].C)





import numpy as np
from itertools import chain, combinations
from shutil import get_terminal_size
from mbnpy.extern.tabulate import tabulate


class Variable:
    """
    A class to manage information about a variable used in matrix-based Bayesian networks.

    Attributes
    ----------
    name : str
        Name of the variable.
    values : list
        Description of basic states.
    B : list of set
        Mapping from basic states to composite states. Each element is a set
        representing a group of basic states (automatically generated).

    Notes
    -----
    **How to read the mapping matrix:**

    Each row in the matrix represents a composite state.
    Each column corresponds to a basic state.

    For example, with 3 basic states, the mapping matrix B is:

        [[1, 0, 0],
         [0, 1, 0],
         [0, 0, 1],
         [1, 1, 0],
         [1, 0, 1],
         [0, 1, 1],
         [1, 1, 1]]

    - The first row `[1, 0, 0]` means that State 0 includes only State 0.
    - The sixth row `[0, 1, 1]` means that State 5 includes States 1 and 2.
    - The last row `[1, 1, 1]` means that State 6 includes all basic states.

    **Best practice when assigning `values`:**

    When applicable, use an ordering where lower indices represent worse outcomes, as some modules assume this ordering.
    For example: `['failure', 'survival']` since `0 < 1`.
    """

    def __init__(self, name: str, values: list=[], B_flag: str='auto'):
        '''Initialise the Variable object.

        Args:
            name (str): name of the variable.
            values (list or np.ndarray): description of states.
            B_flag (str): flag to determine how B is generated.
        '''
        assert isinstance(name, str), 'name should be a string'
        assert isinstance(values, (list, np.ndarray)), \
            'values must be a list or np.ndarray'
        assert B_flag in ['auto', 'store', 'fly'], \
            'B_flag must be either auto, store, or fly'

        self._name = name
        self._values = values
        self._B_flag = B_flag

        if len(self.values) > 0:
            if B_flag == 'fly':
                self._B = None
            elif B_flag == 'store':
                self._B = self.gen_B()
            elif B_flag == 'auto':
                if len(self.values) <= 6:
                    self._B = self.gen_B()
                else:
                    self._B = None
        else:
            self._B = None

    # Magic methods
    def __hash__(self):
        return hash(self._name)

    def __lt__(self, other):
        return self._name < other.name

    def __eq__(self, other):
        if isinstance(other, Variable):
            return self._name == other._name and self._values == other._values
        else:
            return False

    def __repr__(self):
        return (
                f"<Variable representing {self._name}[{', '.join(self._values)}] at {hex(id(self))}>"
           )

    def __str__(self):
        if self._B is None:
            return (
                f"<Variable representing {self._name}[{', '.join(self._values)}] at {hex(id(self))}>\n"
                f"B_flag={repr(self._B_flag)})"
            )
        else:
            header = ['index', 'B']
            b = [(i, f'{x}') for i, x in enumerate(self._B)]
            var_str = tabulate(b, headers=header, tablefmt='grid')
            var_str = self._truncate_strtable(var_str)

            return (
                f"<Variable representing {self._name}[{', '.join(self._values)}] at {hex(id(self))}>\n"
                f"{var_str}"
            )

    def _truncate_strtable(self, var_str):

        terminal_width, terminal_height = get_terminal_size()
        list_rows_str = var_str.split("\n")

        table_width, table_height = len(list_rows_str[0]), len(list_rows_str)

        if table_width > terminal_width:
            colstr_i = np.array(
                [pos for pos, char in enumerate(list_rows_str[0]) if char == "+"]
            )

            half_width = terminal_width // 2 - 3

            left_i = colstr_i[colstr_i < half_width][-1]
            right_i = colstr_i[(table_width - colstr_i) < half_width][0]

            new_var_str = []
            for temp_row_str in list_rows_str:
                left = temp_row_str[: left_i + 1]
                right = temp_row_str[right_i:]
                if temp_row_str[left_i] == "+":
                    joiner = "-----"
                else:
                    joiner = " ... "
                new_var_str.append(left + joiner + right)

            var_str = "\n".join(new_var_str)

        if table_height > terminal_height:

            half_height = terminal_height // 2 - 3

            list_rows_str = var_str.split("\n")
            mid = findnth(list_rows_str[0], '+', 1)

            if half_height % 2:
                first_chunk = half_height + 2
                mid_chunk = half_height + 1
                last_chunk = -half_height + 2
            else:
                first_chunk = half_height + 1
                mid_chunk = half_height
                last_chunk = -half_height + 1

            new_var_str = list_rows_str[:first_chunk]
            no_cross = list_rows_str[mid_chunk].count('+')
            re_line = list(table_width * ' ')
            idx_cross = [findnth(list_rows_str[mid_chunk], '+', i) for i in range(no_cross)]
            mid_cross = [(x + y)//2 for x, y in zip(idx_cross[:-1], idx_cross[1:])]
            re_line = list(table_width * ' ')
            for i in idx_cross:
                re_line[i] = '|'
            for i in mid_cross:
                re_line[i] = ':'
            re_line = ''.join(re_line)
            new_var_str.append(re_line)

            [new_var_str.append(x) for x in list_rows_str[last_chunk:]]

            var_str = "\n".join(new_var_str)

        return var_str

    # Property for 'name'
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        assert isinstance(new_name, str), 'name must be a string'
        self._name = new_name

    # Property for 'values'
    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, new_values):
        assert isinstance(new_values, list), 'values must be a list'
        self._values = new_values
        self.update_B()

    # Property for 'B_flag'
    @property
    def B_flag(self):
        return self._B_flag

    @B_flag.setter
    def B_flag(self, new_B_flag):
        assert new_B_flag in ['auto', 'store', 'fly'], \
            'B_flag must be either auto, store, or fly'
        self._B_flag = new_B_flag
        self.update_B()

    # Method to generate the mapping matrix
    def gen_B(self):
        n = len(self._values)
        B = [
            set(x)
            for x in chain.from_iterable(
                combinations(range(n), r) for r in range(1, n+1)
            )
        ]
        return B

    # Method to update the mapping matrix
    def update_B(self):
        if len(self._values) > 0:
            if self._B_flag == 'store':
                self._B = self.gen_B()
            elif self._B_flag == 'fly':
                self._B = None
            elif self._B_flag == 'auto':
                if len(self._values) <= 6:
                    self._B = self.gen_B()
                else:
                    self._B = None
        else:
            self._B = None

    # Property for 'B'
    @property
    def B(self):
        return self._B


    def get_state(self, state_set):
        '''Finds the state index of a given set of basic states.

        The sets are ordered as follows (cf. gen_B):
        [{0}, {1}, ..., {n-1}, {0, 1}, {0, 2}, ..., {n-2, n-1},
        {0, 1, 2}, ..., {0, 1, ..., n-1}]

        Args:
            state_set (set): set of basic states.

        Returns:
            state (int): state index in B matrix of the given set.
        '''
        assert isinstance(state_set, set), 'set must be a set'

        if self.B is not None:
            # Find the index directly from B
            state = self.B.index(state_set)
        else:
            # Find the index by calculation
            # The number of elements in the target set
            num_elements = len(state_set)
            # Number of basic states
            n = len(self.values)

            # Initialize the state
            state = 0
            # Add the number of sets with fewer elements
            for k in range(1, num_elements):
                state += len(list(combinations(range(n), k)))
            # Find where the target set is in the group
            # with 'num_elements' elements
            combinations_list = list(combinations(range(n), num_elements))

            # Convert target_set to a sorted tuple
            # to match the combinations
            target_tuple = tuple(sorted(state_set))
            # Find the index within the group
            idx_in_group = combinations_list.index(target_tuple)

            # Add the position within the group to the state
            state += idx_in_group

        return state

    def get_set(self, state):
        '''Finds the set of basic states represented by a given state index.

        The sets are ordered as follows (cf. gen_B):
        [{0}, {1}, ..., {n-1}, {0, 1}, {0, 2}, ..., {n-2, n-1},
        {0, 1, 2}, ..., {0, 1, ..., n-1}]

        Args:
            state (int): state index.

        Returns:
            set (set): set of basic states.
        '''
        assert np.issubdtype(type(state), np.integer), 'state must be an integer'

        if self.B is not None:
            return self.B[state]
        else:
            # the number of states
            n = len(self.values)
            # Initialize the state tracker
            current_state = 0

            # Iterate through the group sizes
            # (1-element sets, 2-element sets, etc.)
            for k in range(1, n+1):
                # Count the number of sets of size k
                comb_count = len(list(combinations(range(n), k)))

                # Check if the index falls within this group
                if current_state + comb_count > state:
                    # If it falls within this group,
                    # calculate the exact set
                    combinations_list = list(combinations(range(n), k))
                    set_tuple = combinations_list[state - current_state]
                    return set(set_tuple)

                # Otherwise, move to the next group
                current_state += comb_count

            # If the index is out of bounds, raise an error
            raise IndexError(f"The given state index must be not greater than {2**n-1}")


    def get_state_from_vector(self, vector):
        '''Finds the state index for a given binary vector.

        Args:
            vector (list or np.ndarray): binary vector.
            1 if the basic state is involved, 0 otherwise.

        Returns:
            state (int): state index.
            -1 if the vector is all zeros.
        '''
        assert isinstance(vector, (list, np.ndarray)), \
            'vector must be a list or np.ndarray'
        assert len(vector) == len(self.values), \
            'vector must have the same length as values'

        # Count the number of 1's in the vector
        num_ones = sum(vector)

        # Return -1 if the vector is all zeros
        if num_ones == 0:
            return -1

        # Number of basic states
        n = len(vector)

        # Initialize the state
        state = 0
        # Add the number of vectors with fewer 1's
        for k in range(1, num_ones):
            state += len(list(combinations(range(n), k)))

        # Find where this vector is in the group with 'num_ones' ones
        one_positions = [i for i, val in enumerate(vector) if val == 1]
        # Find the position of this specific combination in the group
        combs = list(combinations(range(n), num_ones))
        idx_in_group = combs.index(tuple(one_positions))

        # Add the position within the group to the state
        state += idx_in_group

        return state

    def get_Bst_from_Bvec( self, Bvec ):
        '''Converts a binary vector into its corresponding state index.

        Args:
           Bvec (np.ndarray): (x*y*z) binary array.
           x is the number of instances for the first Cpm object.
           y is the number of instances for the second Cpm object.
           z is the number of basic states.

        Returns:
            Bst (np.ndarray): (x*y) integer array.
            Each element represents the state index of
            the corresponding binary vector in Bvec.
        '''
        Bst = np.apply_along_axis(self.get_state_from_vector, -1, Bvec)
        return Bst

def findnth(haystack, needle, n):
    parts= haystack.split(needle, n+1)
    if len(parts)<=n+1:
        return -1
    return len(haystack)-len(parts[-1])-len(needle)



import numpy as np
#import dask
import json
import networkx as nx
import socket
import matplotlib
import copy

from mbnpy import variable
from scipy.stats import lognorm

if 'gadi' in socket.gethostname():
    matplotlib.use('Agg')
else:
    matplotlib.use("TKAgg")
    import matplotlib.pyplot as plt


system_meta = {'system_meta': {
        "MODEL_NAME": {
            "value": "Basic Linear Network",
            "notes": None
        },
        "INFRASTRUCTURE_LEVEL": {
            "value": "network",
            "notes": None
        },
        "SYSTEM_CLASS": {
            "value": "RailNetwork",
            "notes": None
        },
        "SYSTEM_SUBCLASS": {
            "value": "Regional Rail Network",
            "notes": None
        },
        "SYSTEM_COMPONENT_LOCATION_CONF": {
            "value": "defined",
            "notes": None
        },
        "RESTORATION_TIME_UNIT": {
            "value": "days",
            "notes": None
        },
        "HAZARD_INTENSITY_MEASURE_PARAM": {
            "value": "PGA",
            "notes": None
        },
        "HAZARD_INTENSITY_MEASURE_UNIT": {
            "value": "g",
            "notes": None
        }
        }
        }


def sys_fun_wrap(G, od_pair, varis=None, thres=None):

    def sys_fun2(comps_st):
        return sf_min_path(comps_st, G, od_pair, varis=varis, thres=thres)
    return sys_fun2


def sf_min_path(comps_st, G, od_pair, varis=None, thres=None):
    """
    comps_st:
    G:
    od_pair:
    varis:
    thres:
    """

    if thres:

        if isinstance(od_pair, (list, tuple)):
            d_time, path_n, path_e = get_time_and_path_given_comps(comps_st, G, od_pair, varis)

        elif isinstance(od_pair, dict): # multiple destinations, e.g. {origin: n1, dests: [n2, n3, n4]}
            d_time, path_n, path_e = get_time_and_path_multi_dest(comps_st, G, od_pair['origin'], od_pair['dests'], varis)

        if d_time > thres:
            sys_st = 'f'
        else:
            sys_st= 's'

    else:
        # no threshold, only check connectivity
        # varis is not used; comps_st has the information (0: fail, 1: success)
        d_time = None

        # comps_st can be either node or edge, or combined
        path_e, path_n = get_connectivity_given_comps(comps_st, G, od_pair)

        if path_n:
            sys_st = 's'
        else:
            sys_st= 'f'

    min_comps_st = {}
    if sys_st == 's':
        # edges
        min_comps_st.update({k: comps_st[k] for k in path_e if k in comps_st})

        # nodes
        min_comps_st.update({k: comps_st[k] for k in path_n if k in comps_st})

    return d_time, sys_st, min_comps_st


def get_connectivity_given_comps(comps_st, G, od_pair):
    """
    return path on connectivity given od_pair including od_pair
    it works on either node or edge
    comps_st: starting from 0: (0: failure, 1: intact) only binary status
              comps_st can be for node, edge or combined
    G: graph
    od_pair:
    arcs:
    vari:
    """
    assert isinstance(comps_st, dict)
    #assert all([v < len(vari[k].values) for k, v in comps_st.items()])

    # node of od_pair should be intact regardless of comps_st
    #_comps_st = comps_st.copy()
    #_comps_st.update({k: 1 for k in od_pair})

    H = update_G_given_comps_st(G, comps_st)

    # shortest path in edges
    try:
        # shortest path in nodes
        s_path = nx.shortest_path(H, source = od_pair[0], target = od_pair[1])

    except (nx.NodeNotFound, nx.exception.NetworkXNoPath):
        s_path, s_path_edge = [], []

    else:
        s_path_edge = get_edge_from_nodes(H, s_path)

    return s_path_edge, s_path


def get_edge_from_nodes(G, path_nodes):

    # path in edges
    path_edges = []
    if isinstance(G, nx.MultiGraph):
        for p in nx.utils.pairwise(path_nodes):
            edges = (G.get_edge_data(*p).items())
            sorted_edges = sorted(edges, key=lambda x: x[1].get('weight'))
            path_edges.append(sorted_edges[0][1]['label'])
    else:
        for p in nx.utils.pairwise(path_nodes):
            path_edges.append(G.get_edge_data(*p)['label'])

    return path_edges


def update_G_given_comps_st(G, comps_st, varis=None):

    assert isinstance(G, nx.Graph), f'G must be an instance of networkx.Graph'

    H = copy.deepcopy(G)

    if varis:

        attributes = {}
        if isinstance(H, nx.MultiGraph):
            for e in H.edges(data=True):
                k = e[2]['label']
                attributes[(e[0], e[1], 0)] = {'weight': varis[k].values[comps_st[k]]}
        else:
            for e in H.edges(data=True):
                k = e[2]['label']
                attributes[(e[0], e[1])] = {'weight': varis[k].values[comps_st[k]]}

        # update H
        nx.set_edge_attributes(H, attributes)

    else:

        # edge
        for n1, n2, d in list(H.edges(data=True)):
            try:
                s = comps_st[d['label']]
            except KeyError:
                pass
            else:
                if not s:
                    try:
                        H.remove_edge(n1, n2, key=d['label'])
                    except TypeError:
                        H.remove_edge(n1, n2)

        # node
        for n0 in list(H.nodes):
            try:
                s = comps_st[n0]
            except KeyError:
                pass
            else:
                if not s:
                    H.remove_node(n0)

    return H


def get_time_and_path_given_comps(comps_st, G, od_pair, varis):
    """
    comps_st: starting from 0
    od_pair:
    arcs:
    vari:
    """
    assert isinstance(comps_st, dict)
    assert all([comps_st[k] < len(v.values) for k, v in varis.items()])

    H = update_G_given_comps_st(G, comps_st, varis)

    path = nx.shortest_path(H, source=od_pair[0], target=od_pair[1], weight='weight')
    d_time = nx.shortest_path_length(H, source=od_pair[0], target=od_pair[1], weight='weight')
    path_e = get_edge_from_nodes(H, path)

    return d_time, path, path_e


def get_time_and_path_multi_dest(comps_st, G, origin, dests, varis):
    """
    Compute time and path given multiple destinations--NB: works only for bidirectional graph
    """
    assert isinstance(dests, (list, tuple)), f'dests must be a list-like: {type(dests)}'
    assert isinstance(comps_st, dict)
    assert all([comps_st[k] < len(varis[k].values) for k in comps_st.keys()])

    # G must be a undirected (either Graph or MultiGraph)
    assert not isinstance(G, (nx.DiGraph, nx.MultiDiGraph)), f'G must not be a directed graph'

    H = update_G_given_comps_st(G, comps_st, varis)

    try:
        d_time, path_nodes = nx.multi_source_dijkstra(H, sources=dests, target=origin, weight='weight')
    except nx.exception.NetworkXNoPath:
        d_time, path_nodes, path_edges = np.inf, [], []
    else:
        path_nodes = path_nodes[::-1] # reverse as origin was used as the target
        path_edges = get_edge_from_nodes(H, path_nodes)

    return d_time, path_nodes, path_edges


def get_all_paths_and_times(ODs, G, key='weight'):
    """
    ODs: list of OD pairs
    G: instance of networkx.Graph
    key: keyword for weight
    """

    path_time = {}
    for org, dest in ODs:
        for _path in nx.all_simple_paths(G, org, dest):
            val = nx.path_weight(G, _path, weight=key)

            edges_path = []
            for x in list(zip(_path, _path[1:])):
                edges_path.append(G[x[0]][x[1]]['label'])

            path_time.setdefault((org, dest), []).append((edges_path, val))

    return path_time


def get_path_time_idx(path_time, vari):
    """
    path_time: a list of tuple
    vari: instance of variable.Variable

    """
    assert isinstance(path_time, list)
    assert all([isinstance(x, tuple) for x in path_time])
    assert all([len(x)==2 for x in path_time])
    assert isinstance(vari, variable.Variable)

    path_timex = path_time[:]

    # refering variable
    path_time_idx = []
    for x in path_timex:
        idx = [i for i, y in enumerate(vari.values) if np.isclose(x[1], y)]
        try:
            path_time_idx.append((*x, idx[0]))
        except IndexError:
            print('path_time incompatible with variable')

    # sort by elapsed time
    path_time_idx = sorted(path_time_idx, key=lambda x: x[2], reverse=True)

    if not any([np.inf in x for x in path_timex]):
        path_time_idx.insert(0, ([], np.inf, 0))

    return path_time_idx


def get_arcs_length(arcs, node_coords):
    """
    if arcs.shape[1] == 2:
        nArc = arcs.shape[0]
    elif arcs.shape[0] == 2:
        arcs = arcs.T
        nArc = arcs.shape[0]
    else:
        print('"arcs" must have either two columns or two rows (each noting start and end points)')

    if node_coords.shape[1] != 2:
        if node_coords.shape[0] == 2:
            node_coords = nodeCoords.T
        else:
            print('"node_coords" must have either two columns or two rows (each noting coordinates of x and y).')
    """
    arc_len = {}
    for k, v in arcs.items():
        diff = np.array(node_coords[v[0]]) - np.array(node_coords[v[1]])
        arc_len[k] = np.sqrt(np.sum(diff**2))

    return arc_len


def eval_sys_state_given_arc(arcs_state, **kwargs):
    """
    arcs_state: dict or frozenset
    path_time_idx: a list of tuple (path, time, idx)
    arc_cond: value for survival (row index)
    """

    path_time_idx = kwargs['path_time_idx']
    arc_cond = kwargs['arc_cond']

    sys_state = path_time_idx[0][2]  # no path available

    for edges, _, state in path_time_idx:

        path_is_surv = [arcs_state[i]==arc_cond for i in edges]
        if path_is_surv and all(path_is_surv):
            sys_state = state
            break

    return sys_state


def eval_sys_state(path_time_idx, arcs_state, arc_cond):
    """
    path_time_idx: a list of tuple (path, time, idx)
    arcs_state: dict or frozenset
    arc_cond: value for survival (row index)
    """

    sys_state = path_time_idx[0][2]  # no path available

    for edges, _, state in path_time_idx:

        path_is_surv = [arcs_state[i]==arc_cond for i in edges]
        if path_is_surv and all(path_is_surv):
            sys_state = state
            break

    return sys_state


def get_node_conn_df(arcs, node_coords, avg_speed_by_arc, outfile=None):

    distance_by_arc = get_arcs_length(arcs, node_coords)

    #arcTimes_h = arcLens_km ./ arcs_Vavg_kmh
    _dic = {'node_conn_df': {}}
    for k, v in arcs.items():

        _dic['node_conn_df'][k] = {'origin': v[0],
                   'destination': v[1],
                   'link_capacity': None,
                   'weight': distance_by_arc[k]/avg_speed_by_arc[k]}

    if outfile:
        with open(outfile, 'w') as w:
            json.dump(_dic, w, indent=4)
        print(f'{outfile} is written')

    return _dic


def create_model_json_for_graph_network(arcs, node_coords, ODs, outfile=None):

    _dic = {}
    _dic.update(system_meta)

    sysout_setup = {'sysout_setup': {}}
    for k, v in ODs.items():
        sysout_setup['sysout_setup'][k] = {'origin': v[0],
                           'destination': v[1],
                           'output_node_capacity': None,
                           'capacity_fraction': None,
                           'priorty': None
                           }
    _dic.update(sysout_setup)

    #arcTimes_h = arcLens_km ./ arcs_Vavg_kmh
    _dic['node_conn_df'] = {}
    for k, v in arcs.items():
        _dic['node_conn_df'][k] = {'origin': v[0],
                   'destination': v[1],
                   'link_capacity': None,
                   'weight': 1.0}

    component_list = {'component_list': {}}
    for k, v in node_coords.items():
        component_list['component_list'][k] = {'component_type': None,
                             'component_class': None,
                             'cost_fraction': None,
                             'cost_fraction': None,
                             'node_type': None,
                             'node_cluster': None,
                             'operation_capacity': None,
                             'pos_x': v[0],
                             'pos_y': v[1],
                             'damages_states_constructor': None}
    _dic.update(component_list)

    if outfile:
        with open(outfile, 'w') as w:
            json.dump(_dic, w, indent=4)
        print(f'{outfile} is written')

    return _dic


def create_model_json_for_tranportation_network(arcs, node_coords, avg_speed_by_arc, ODs, outfile=None):

    _dic = {}
    _dic.update(system_meta)

    sysout_setup = {'sysout_setup': {}}
    for k, v in ODs.items():
        sysout_setup['sysout_setup'][k] = {'origin': v[0],
                           'destination': v[1],
                           'output_node_capacity': None,
                           'capacity_fraction': None,
                           'priorty': None
                           }
    _dic.update(sysout_setup)
    node_conn_df = get_node_conn_df(arcs, node_coords, avg_speed_by_arc)
    _dic.update(node_conn_df)

    component_list = {'component_list': {}}
    for k, v in node_coords.items():
        component_list['component_list'][k] = {'component_type': None,
                             'component_class': None,
                             'cost_fraction': None,
                             'cost_fraction': None,
                             'node_type': None,
                             'node_cluster': None,
                             'operation_capacity': None,
                             'pos_x': v[0],
                             'pos_y': v[1],
                             'damages_states_constructor': None}
    _dic.update(component_list)

    if outfile:
        with open(outfile, 'w') as w:
            json.dump(_dic, w, indent=4)
        print(f'{outfile} is written')

    return _dic


def create_scenario_json_for_trans_network(damage_states, scenarios_dic, outfile=None):

    assert isinstance(damage_states, list), 'damage_states should be a list'
    assert isinstance(scenarios_dic, dict), 'scenarios_dic should be a dict'

    _dic = {'damage_states': damage_states}

    scenarios = {'scenarios': {}}
    for k, v in scenarios_dic.items():

        scenarios['scenarios'][k] = v

    _dic.update(scenarios)

    if outfile:
        with open(outfile, 'w') as w:
            json.dump(_dic, w, indent=4)
        print(f'{outfile} is written')

    return _dic


def create_scenario_json_for_tranportation_network(damage_states, arcs, type_by_arc, frag_by_type, obs_by_arc, key=None):
    """
    damage_states: list of string
    arcs: dict
    type_by_arc: dict
    frag_by_type: dict (only support lognorm.cdf, 'std', 'med')
    obs_by_arc: dict
    """
    _dic = {'damage_states': damage_states}
    s1_list = {'scenarios': {'s1': {}}}

    for k in arcs.keys():

        _type = type_by_arc[k]

        prob = lognorm.cdf(obs_by_arc[k],
                           frag_by_type[_type]['std'],
                           scale=frag_by_type[_type]['med'])

        s1_list['scenarios']['s1'][k] = [prob, 1-prob]

    _dic.update(s1_list)

    if key:
        wfile = f'./scenarios_{key}.json'
        with open(wfile, 'w') as w:
            json.dump(_dic, w, indent=4)
        print(f'{wfile} is written')

    return _dic


def plot_graph(G, filename=None):

    # plot graph
    pos = nx.get_node_attributes(G, 'pos')
    edge_labels = nx.get_edge_attributes(G, 'label')
    fig = plt.figure()
    ax = fig.add_subplot()
    nx.draw(G, pos, with_labels=True, ax=ax)
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, ax=ax)

    if filename:
       fig.savefig(filename, dpi=200)
       print(f'{filename} is created')

